# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from pyfdb.pyfdb import FDB
from pyfdb.pyfdb_iterator import StatusElement


def test_status(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
    }

    status_iterator = fdb.status(selection)

    elements = []

    for el in status_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 32

    expected_key = {
        "class": ["ea"],
        "date": ["20200104"],
        "domain": ["g"],
        "expver": ["0001"],
        "stream": ["oper"],
    }
    expected_control_identifiers: list[StatusElement] = []

    for element in elements:
        assert all(
            el in element.key() for el in expected_key
        )  # Check whether the expected key is part of the elements

        assert element.controlIdentifiers() == expected_control_identifiers

        date = element.key()["date"]
        assert date and len(date) == 1
        time = element.key()["time"]
        assert time and len(time) == 1

        assert "ea:0001:oper:" in str(element.location())
        assert f":{date[0]}:" in str(element.location())
        assert f":{time[0]}:" in str(element.location())
        assert ":g" in str(element.location())

    # Test compare of status elements
    assert elements[0] == elements[0]
    assert elements[1] != elements[0]



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.

import enum
import pathlib

import chunked_data_view_bindings.chunked_data_view_bindings as pdv

from pychunked_data_view.exceptions import MarsRequestFormattingError, InternalError

pdv.init_bindings()


class Chunking(enum.Enum):
    """Defines how a axis will be chunked

    Attributes:
        NONE: Axis will not be chunked, accessing any value from this axis will load all values.
        SINGLE_VALUE: Axis will be chunked. One chunk per value.
    """

    NONE = enum.auto()
    SINGLE_VALUE = enum.auto()


class AxisDefinition:
    @staticmethod
    def _translate_chunking(
        chunking: Chunking,
    ) -> pdv.AxisDefinition.NoChunking | pdv.AxisDefinition.IndividualChunking:
        if chunking == Chunking.NONE:
            return pdv.AxisDefinition.NoChunking()
        elif chunking == Chunking.SINGLE_VALUE:
            return pdv.AxisDefinition.IndividualChunking()
        else:
            raise InternalError()

    def __init__(self, keys: list[str], chunking: Chunking):
        """Defines which axis from a MARS Request form an axis in the Zarr array.

        Also defines if the data is to be chunked.

        Args:
            keys(list of str): mars keys that for this axis.
            chunking ( Chunking): Define how this axis shall be chunked
        """
        self._obj = pdv.AxisDefinition(
            keys=keys, chunking=self._translate_chunking(chunking)
        )

    @property
    def keys(self) -> list[str]:
        return self._obj.keys

    @keys.setter
    def keys(self, keys: list[str]) -> None:
        self._obj.keys = keys

    @property
    def chunking(self) -> Chunking:
        chunking = self._obj.chunking
        if isinstance(chunking, pdv.AxisDefinition.NoChunking):
            return Chunking.NONE
        elif isinstance(chunking, pdv.AxisDefinition.IndividualChunking):
            return Chunking.SINGLE_VALUE
        else:
            raise InternalError()

    @chunking.setter
    def chunking(self, chunking: Chunking) -> None:
        self._obj.chunking = self._translate_chunking(chunking)


class ChunkedDataView:
    def __init__(self, obj: pdv.ChunkedDataView):
        self._obj = obj

    def at(self, index: list[int] | tuple[int, ...]):
        return self._obj.at(index)

    def chunkShape(self):
        return self._obj.chunk_shape()

    def chunks(self):
        return self._obj.chunks()

    def shape(self):
        return self._obj.shape()


class ExtractorType(enum.Enum):
    """Suported data extractors.

    Defines what storage format the caller expects to be stored in FDB.
    """

    GRIB = pdv.ExtractorType.GRIB
    """Extract data from GRIB"""


class ChunkedDataViewBuilder:
    def __init__(self, fdb_config_file: pathlib.Path | None):
        self._obj = pdv.ChunkedDataViewBuilder(fdb_config_file)

    def add_part(
        self,
        mars_request_key_values: str,
        axes: list[AxisDefinition],
        extractor_type: ExtractorType,
    ):
        self._obj.add_part(
            mars_request_key_values, [ax._obj for ax in axes], extractor_type.value
        )

    def extend_on_axis(self, axis: int):
        self._obj.extend_on_axis(axis)

    def build(self):
        try:
            return ChunkedDataView(self._obj.build())
        except RuntimeError as re:
            exception_msg = str(re)
            if "StreamParser::next" in exception_msg:
                raise MarsRequestFormattingError(
                    exception_msg + "\n Did the MARS request end in a comma?"
                )
            elif "MarsParser::parseVerb" in exception_msg:
                raise MarsRequestFormattingError(
                    exception_msg + "\n Did you miss a comma between keys?"
                )
            elif "Cannot match" in exception_msg:
                raise MarsRequestFormattingError(
                    exception_msg + "\n Did you misspell a MARS key?"
                )
            else:
                raise re



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from pyfdb import FDB


def test_purge_dryrun_no_purge(empty_fdb_setup, build_grib_messages):
    fdb = FDB(empty_fdb_setup)

    initial_elements = list(fdb.list({"class": "ea"}))

    assert len(initial_elements) == 0
    print(f"Elements initial: {len(initial_elements)}")

    fdb.archive(build_grib_messages.read_bytes())
    fdb.flush()

    print("Archiving again")
    fdb.archive(build_grib_messages.read_bytes())
    fdb.flush()

    after_archive_elements = list(fdb.list({"class": "ea"}))

    print(
        f"Elements after masking archive: {len(after_archive_elements)}, before: {len(initial_elements)}"
    )
    assert len(after_archive_elements) > len(initial_elements)

    purge_iterator = fdb.purge({"class": "ea"})
    purged_output = list(purge_iterator)

    assert len(purged_output) > 0

    elements_after_second_purge = list(fdb.list({"class": "ea"}))

    for el in elements_after_second_purge:
        print(el)

    print(f"Elements after second purge: {len(elements_after_second_purge)}")
    assert len(after_archive_elements) == len(elements_after_second_purge)


def test_purge(empty_fdb_setup, build_grib_messages):
    fdb = FDB(empty_fdb_setup)

    initial_elements = list(fdb.list({"class": "ea"}))

    assert len(initial_elements) == 0
    print(f"Elements initial: {len(initial_elements)}")

    fdb.archive(build_grib_messages.read_bytes())
    fdb.flush()

    archived_elements = list(fdb.list({"class": "ea"}))
    assert len(archived_elements) == 96

    print(f"Elements after archive: {len(archived_elements)}")

    purge_iterator = fdb.purge({"class": "ea"}, doit=True)
    purged_output = []

    for el in purge_iterator:
        print(el)
        purged_output.append(el)

    assert len(purged_output) > 0

    elements_after_first_purge = list(fdb.list({"class": "ea"}))

    print(f"Elements after purge: {len(elements_after_first_purge)}")
    assert len(archived_elements) == len(elements_after_first_purge)

    fdb.archive(build_grib_messages.read_bytes())
    fdb.flush()

    after_archive_elements = list(fdb.list({"class": "ea"}))

    print(
        f"Elements after masking archive: {len(after_archive_elements)}, before: {len(initial_elements)}"
    )
    assert len(after_archive_elements) > len(initial_elements)

    purge_iterator = fdb.purge({"class": "ea"}, doit=True)
    purged_output = list(purge_iterator)

    assert len(purged_output) > 0

    elements_after_second_purge = list(fdb.list({"class": "ea"}))

    for el in elements_after_second_purge:
        print(el)

    print(f"Elements after second purge: {len(elements_after_second_purge)}")
    assert len(archived_elements) == len(elements_after_second_purge)



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
import datetime

project = "PyFDB"
copyright = f"{datetime.datetime.today().year}, ECMWF"
author = "ECMWF"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.autosectionlabel",
    "sphinxcontrib.mermaid",
    "autoapi.extension",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.inheritance_diagram",
]

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store", "_internal"]

autoapi_dirs = ["../../src/pyfdb"]
autoapi_type = "python"
autoapi_generate_api_docs = True
autoapi_add_toctree_entry = False
autoapi_python_class_content = "class"
autoapi_ignore = [
    "*/_internal/*",
]
add_module_names = False
autoapi_keep_files = False

# -- Napoleon settings
napoleon_google = True

templates_path = ["_templates"]
exclude_patterns = ["Thumbs.db", ".DS_Store"]


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "pydata_sphinx_theme"
# Do not show source link on each page
html_show_sourcelink = False
html_sidebars = {
    "**": [],
}
html_theme_options = {
    "navbar_align": "left",
    "navbar_start": ["navbar-logo"],
    "navbar_center": ["navbar-nav"],
    "navbar_end": ["navbar-icon-links", "theme-switcher", "version-switcher"],
    "navbar_persistent": ["search-button"],
    "primary_sidebar_end": [],
    # On local builds no version.json is present
    "check_switcher": False,
}

html_context = {
    # Enable auto detection of light/dark mode
    "default_mode": "auto"
}

# -- autosectionlabel configuration ------------------------------------------
autosectionlabel_prefix_document = True



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from pyfdb import FDB
from pyfdb.pyfdb_iterator import WipeElementType, _WipeElementType


def test_wipe_enum_mapping():
    assert WipeElementType.ERROR.value == _WipeElementType.ERROR.value
    assert WipeElementType.CATALOGUE_INFO.value == _WipeElementType.CATALOGUE_INFO.value
    assert WipeElementType.CATALOGUE.value == _WipeElementType.CATALOGUE.value
    assert (
        WipeElementType.CATALOGUE_INDEX.value == _WipeElementType.CATALOGUE_INDEX.value
    )
    assert WipeElementType.CATALOGUE_SAFE.value == _WipeElementType.CATALOGUE_SAFE.value
    assert (
        WipeElementType.CATALOGUE_CONTROL.value
        == _WipeElementType.CATALOGUE_CONTROL.value
    )
    assert WipeElementType.STORE.value == _WipeElementType.STORE.value
    assert WipeElementType.STORE_AUX.value == _WipeElementType.STORE_AUX.value
    assert WipeElementType.STORE_SAFE.value == _WipeElementType.STORE_SAFE.value
    assert WipeElementType.UNKNOWN.value == _WipeElementType.UNKNOWN.value


def test_wipe_dryrun(read_write_fdb_setup):
    fdb = FDB(read_write_fdb_setup)

    elements = list(fdb.list({"class": "ea"}))
    assert len(elements) > 0

    wipe_iterator = fdb.wipe({"class": "ea"})
    wiped_elements = list(wipe_iterator)
    assert len(wiped_elements) > 0

    for el in wiped_elements:
        print(el)

    elements_after_wipe = list(fdb.list({"class": "ea"}))
    assert len(elements) == len(elements_after_wipe)


def test_wipe_all_doit(read_write_fdb_setup):
    fdb = FDB(read_write_fdb_setup)

    elements = list(fdb.list({"class": "ea"}))
    assert len(elements) > 0

    wipe_iterator = fdb.wipe({"class": "ea"}, doit=True)
    wiped_elements = list(wipe_iterator)
    assert len(wiped_elements) > 0

    elements_after_wipe = list(fdb.list({"class": "ea"}))
    print(
        f"#Elements before: {len(elements)}, Elements after: {len(elements_after_wipe)}"
    )
    assert len(elements) > len(elements_after_wipe)


def test_wipe_single_date_doit(read_write_fdb_setup):
    fdb = FDB(read_write_fdb_setup)

    elements = list(fdb.list({"class": "ea"}))
    assert len(elements) > 0

    wipe_iterator = fdb.wipe({"class": "ea", "date": "20200101"}, doit=True)
    wiped_elements = list(wipe_iterator)
    assert len(wiped_elements) > 0

    elements_after_wipe = list(fdb.list({"class": "ea"}))
    print(
        f"#Elements before: {len(elements)}, Elements after: {len(elements_after_wipe)}"
    )
    assert len(elements) > len(elements_after_wipe)
    assert len(elements) == 96
    assert len(elements_after_wipe) == 72


def populate_fdb(fdb: FDB):
    selection = {
        "class": "rd",
        "expver": "xxxx",
        "stream": "oper",
        "type": "fc",
        "date": "20000101",
        "time": "0000",
        "domain": "g",
        "levtype": "pl",
        "levelist": "300",
        "param": "138",
        "step": "0",
    }

    # Write 4 fields to the FDB based on BASE_REQUEST
    requests = [selection.copy() for _ in range(4)]

    # Modify on each of the 3 levels of the schema
    requests[1]["step"] = "1"
    requests[2]["date"] = "20000102"
    requests[3]["levtype"] = "sfc"
    del requests[3]["levelist"]

    number_of_fields = 4

    data = b"-1 Kelvin"
    for i in range(number_of_fields):
        key = requests[i]
        key = [(k, v) for k, v in key.items()]
        fdb.archive(identifier=key, data=data)
    fdb.flush()

    return number_of_fields


def test_wipe_list(empty_fdb_setup):
    fdb = FDB(empty_fdb_setup)

    number_of_fields = populate_fdb(fdb)
    assert len(list(fdb.list({}))) == number_of_fields

    # Wipe without doit: Do not actually delete anything.
    wipe_iterator = fdb.wipe({"class": "rd"})

    assert len(list(fdb.list({}))) == number_of_fields

    # Wipe, do it
    wipe_iterator = fdb.wipe({"class": "rd"}, doit=True)

    # Consume all wipe iterator elements
    for el in wipe_iterator:
        print(el.msg())
        print(el.type())
        print(el.uris())

    elements = list(fdb.list({}))
    assert len(elements) == 0



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest

from pyfdb import FDB


def test_list_empty_fdb_list_non_existing(empty_fdb_setup):
    fdb = FDB(empty_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "time": "1800",
    }

    list_iterator = fdb.list(selection, level=3)
    assert list_iterator

    elements = list(list_iterator)
    assert len(elements) == 0


def test_list_empty_fdb_list_all(empty_fdb_setup):
    fdb = FDB(empty_fdb_setup)

    list_iterator = fdb.list({}, level=3)
    assert list_iterator

    elements = list(list_iterator)
    assert len(elements) == 0


def test_list_partial(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {"type": "an"}

    list_iterator = fdb.list(selection, level=1)
    assert list_iterator

    elements = list(list_iterator)
    assert len(elements) == 32


@pytest.mark.parametrize("level, expected_key_elements", [(3, 10), (2, 8), (1, 6)])
def test_list_partial_contains_combined_key(read_only_fdb_setup, level, expected_key_elements):
    fdb = FDB(read_only_fdb_setup)

    selection = {"type": "an"}

    elements = list(fdb.list(selection, level=level))

    for element in elements:
        print(
            f"Element Offset: {element.offset()}, Element Length: {element.length()}, Element Key: {element.combined_key()}"
        )
        combined_key = element.combined_key()

        if level == 3 or level == 2:
            assert "type" in combined_key
            assert combined_key["type"] == "an"
        else:
            assert "type" not in combined_key

        if level == 3:
            assert element.offset() is not None
            assert element.length() is not None
        else:
            assert element.offset() is None
            assert element.length() is None

        assert len(combined_key) == expected_key_elements


@pytest.mark.parametrize("level", [3, 2, 1])
def test_list_partial_contains_keys(read_only_fdb_setup, level):
    fdb = FDB(read_only_fdb_setup)

    selection = {"type": "an"}

    elements = list(fdb.list(selection, level=level))

    for element in elements:
        print(f"Element Offset: {element.offset()}, Element Length: {element.length()}, Element Key: {element.keys()}")
        keys = element.keys()

        # type is part of the second level
        if level == 3:
            assert "type" in element.keys()[1]
            assert keys[1]["type"] == "an"
            assert len(keys[0]) == 6
            assert len(keys[1]) == 2
            assert len(keys[2]) == 2
            assert element.offset() is not None
            assert element.length() is not None
        elif level == 2:
            assert "type" in element.keys()[1]
            assert keys[1]["type"] == "an"
            assert len(keys[0]) == 6
            assert len(keys[1]) == 2
            assert keys[2] == {}
            assert element.offset() is None
            assert element.length() is None
        else:
            assert len(element.keys()[0]) == 6
            assert keys[1] == {}
            assert keys[2] == {}
            assert element.offset() is None
            assert element.length() is None


def test_read_only_attributes_list_element(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "time": "1800",
    }

    list_iterator = fdb.list(selection, level=3)
    assert list_iterator

    elements = list(list_iterator)

    for el in elements:
        with pytest.raises(AttributeError):
            el.data_handle = None

    for el in elements:
        with pytest.raises(AttributeError):
            el.uri = None


def test_list_minimal(read_only_fdb_setup, test_logger):
    with FDB(read_only_fdb_setup) as fdb:
        selection = {
            "type": "an",
            "class": "ea",
            "domain": "g",
            "expver": "0001",
            "stream": "oper",
            "date": "20200101",
            "levtype": "sfc",
            "step": "0",
            "time": "1800",
        }
        test_logger.debug(f"Stringified selection:\n  {selection}")

        list_iterator = fdb.list(selection, level=1)
        assert list_iterator

        elements = []

        for el in list_iterator:
            test_logger.debug(el)
            elements.append(el)
            test_logger.debug(f"Has Location: {el.has_location()}")

        assert len(elements) == 1

        test_logger.debug("----------------------------------")

        selection = {
            "type": "an",
            "class": "ea",
            "domain": "g",
            "expver": "0001",
            "stream": "oper",
            "date": "20200101",
            "levtype": "sfc",
            "step": "0",
            "time": "1800",
        }

        list_iterator = fdb.list(selection, level=2)
        assert list_iterator

        elements = []

        for el in list_iterator:
            test_logger.debug(el)
            elements.append(el)
            test_logger.debug(f"Has Location: {el.has_location()}")

        assert len(elements) == 1

        test_logger.debug("----------------------------------")

        selection = {
            "type": "an",
            "class": "ea",
            "domain": "g",
            "expver": "0001",
            "stream": "oper",
            "date": "20200101",
            "levtype": "sfc",
            "step": "0",
            "time": "1800",
        }

        list_iterator = fdb.list(selection, level=3)
        assert list_iterator

        elements = []

        for el in list_iterator:
            test_logger.debug(el)
            test_logger.debug(f"Has Location: {el.has_location()}")
            test_logger.debug(f"Has URI: {el.uri}")
            test_logger.debug(f"Has Offset: {el.offset}")
            test_logger.debug(f"Has Length: {el.length}")
            elements.append(el)

        assert len(elements) == 3


def test_list_deduplicate(read_only_fdb_setup, build_grib_messages):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "167",
        "time": "1800",
    }

    list_iterator = fdb.list(selection, include_masked=False, level=3)
    assert list_iterator

    elements = []

    for el in list_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 1

    print("Archiving duplicate data")
    fdb.archive(build_grib_messages.read_bytes())
    fdb.flush()

    list_iterator = fdb.list(selection, include_masked=True, level=3)
    assert list_iterator
    elements = []

    for el in list_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 2


def test_list_read_from_datahandle(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }
    print(f"Stringified selection:\n  {selection}")

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "time": "1800",
    }

    list_iterator = fdb.list(selection, level=3)
    assert list_iterator

    elements = []

    for el in list_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 3

    for el in elements:
        data_handle = el.data_handle
        data_handle.open()
        assert data_handle.read(4) == b"GRIB"
        data_handle.close()



import zarr
import numpy as np

from z3fdb import (
    AxisDefinition,
    Chunking,
    ExtractorType,
    SimpleStoreBuilder,
)


def test_axis_check_merge(read_only_fdb_setup_for_sfc_pl_example) -> None:
    """This test checks whether an access to an axis which has no pendant in the data is failing.
    The request below has param 167 which is not given in the data of the setup fdb. Therefore this
    needs to fail. Accessing the first two params is fine.
    """
    builder = SimpleStoreBuilder(read_only_fdb_setup_for_sfc_pl_example)
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/2020-01-02,"
        "levtype=sfc,"
        "step=0,"
        "param=165/166,"
        "time=0/to/21/by/3",
        [
            AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/2020-01-02,"
        "levtype=pl,"
        "step=0,"
        "param=131/132,"
        "levelist=50/100,"
        "time=0/to/21/by/3",
        [
            AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param", "levelist"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    builder.extend_on_axis(1)
    store = builder.build()

    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)

    assert data

    assert np.all(data[0, 0] == 0)   # date=20200101 time=0, 10u (sfc 165)
    assert np.all(data[0, 1] == 1)   # date=20200101 time=0, 10v (sfc 166)
    assert np.all(data[0, 2] == 31)  # date=20200101 time=0, u (pl 131) level=50
    assert np.all(data[0, 3] == 33)  # date=20200101 time=0, u (pl 131) level=100
    assert np.all(data[0, 4] == 32)  # date=20200101 time=0, v (pl 132) level=50
    assert np.all(data[0, 5] == 34)  # date=20200101 time=0, v (pl 132) level=100


def test_axis_check_merge_no_chunking(read_only_fdb_setup_for_sfc_pl_example) -> None:
    """This test checks whether an access to an axis which has no pendant in the data is failing.
    The request below has param 167 which is not given in the data of the setup fdb. Therefore this
    needs to fail. Accessing the first two params is fine.
    """
    builder = SimpleStoreBuilder(read_only_fdb_setup_for_sfc_pl_example)
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/2020-01-02,"
        "levtype=sfc,"
        "step=0,"
        "param=165/166,"
        "time=0/to/21/by/3",
        [AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE), AxisDefinition(["param"], Chunking.NONE)],
        ExtractorType.GRIB,
    )
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/2020-01-02,"
        "levtype=pl,"
        "step=0,"
        "param=131/132,"
        "levelist=50/100,"
        "time=0/to/21/by/3",
        [
            AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param", "levelist"], Chunking.NONE),
        ],
        ExtractorType.GRIB,
    )
    builder.extend_on_axis(1)
    store = builder.build()

    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)

    assert data

    assert np.all(data[0, 0] == 0)   # date=20200101 time=0, 10u (sfc 165)
    assert np.all(data[0, 1] == 1)   # date=20200101 time=0, 10v (sfc 166)
    assert np.all(data[0, 2] == 31)  # date=20200101 time=0, u (pl 131) level=50
    assert np.all(data[0, 3] == 33)  # date=20200101 time=0, u (pl 131) level=100
    assert np.all(data[0, 4] == 32)  # date=20200101 time=0, v (pl 132) level=50
    assert np.all(data[0, 5] == 34)  # date=20200101 time=0, v (pl 132) level=100


SFC_REQUEST = (
    "type=an,"
    "class=ea,"
    "domain=g,"
    "expver=0001,"
    "stream=oper,"
    "date=2020-01-01/2020-01-02,"
    "levtype=sfc,"
    "step=0,"
    "param=165/166,"
    "time=0/to/21/by/3"
)

PL_REQUEST = (
    "type=an,"
    "class=ea,"
    "domain=g,"
    "expver=0001,"
    "stream=oper,"
    "date=2020-01-01/2020-01-02,"
    "levtype=pl,"
    "step=0,"
    "param=131/132,"
    "levelist=50/100,"
    "time=0/to/21/by/3"
)


def test_extend_on_axis_0(read_only_fdb_setup_for_sfc_pl_example) -> None:
    """Extension on axis 0 with SINGLE_VALUE chunking. Param axis before date+time."""
    builder = SimpleStoreBuilder(read_only_fdb_setup_for_sfc_pl_example)
    builder.add_part(
        SFC_REQUEST,
        [AxisDefinition(["param"], Chunking.SINGLE_VALUE), AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE)],
        ExtractorType.GRIB,
    )
    builder.add_part(
        PL_REQUEST,
        [
            AxisDefinition(["param", "levelist"], Chunking.SINGLE_VALUE),
            AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    builder.extend_on_axis(0)
    store = builder.build()

    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)

    assert data

    # data[param_idx, date_time_idx]
    assert np.all(data[0, 0] == 0)   # 10u (sfc 165), date=20200101 time=0
    assert np.all(data[1, 0] == 1)   # 10v (sfc 166), date=20200101 time=0
    assert np.all(data[2, 0] == 31)  # u (pl 131) level=50, date=20200101 time=0
    assert np.all(data[3, 0] == 33)  # u (pl 131) level=100, date=20200101 time=0
    assert np.all(data[4, 0] == 32)  # v (pl 132) level=50, date=20200101 time=0
    assert np.all(data[5, 0] == 34)  # v (pl 132) level=100, date=20200101 time=0


def test_extend_on_axis_0_no_chunking(read_only_fdb_setup_for_sfc_pl_example) -> None:
    """Extension on axis 0 with NONE chunking on extension axis."""
    builder = SimpleStoreBuilder(read_only_fdb_setup_for_sfc_pl_example)
    builder.add_part(
        SFC_REQUEST,
        [AxisDefinition(["param"], Chunking.NONE), AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE)],
        ExtractorType.GRIB,
    )
    builder.add_part(
        PL_REQUEST,
        [
            AxisDefinition(["param", "levelist"], Chunking.NONE),
            AxisDefinition(["date", "time"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    builder.extend_on_axis(0)
    store = builder.build()

    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)

    assert data

    # data[param_idx, date_time_idx]
    assert np.all(data[0, 0] == 0)   # 10u (sfc 165), date=20200101 time=0
    assert np.all(data[1, 0] == 1)   # 10v (sfc 166), date=20200101 time=0
    assert np.all(data[2, 0] == 31)  # u (pl 131) level=50, date=20200101 time=0
    assert np.all(data[3, 0] == 33)  # u (pl 131) level=100, date=20200101 time=0
    assert np.all(data[4, 0] == 32)  # v (pl 132) level=50, date=20200101 time=0
    assert np.all(data[5, 0] == 34)  # v (pl 132) level=100, date=20200101 time=0


def test_non_extension_axis_no_chunking(read_only_fdb_setup_for_sfc_pl_example) -> None:
    """NONE chunking on non-extension axis (date+time), SINGLE_VALUE on extension axis (param)."""
    builder = SimpleStoreBuilder(read_only_fdb_setup_for_sfc_pl_example)
    builder.add_part(
        SFC_REQUEST,
        [AxisDefinition(["date", "time"], Chunking.NONE), AxisDefinition(["param"], Chunking.SINGLE_VALUE)],
        ExtractorType.GRIB,
    )
    builder.add_part(
        PL_REQUEST,
        [
            AxisDefinition(["date", "time"], Chunking.NONE),
            AxisDefinition(["param", "levelist"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    builder.extend_on_axis(1)
    store = builder.build()

    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)

    assert data

    # data[date_time_idx, param_idx]
    assert np.all(data[0, 0] == 0)   # date=20200101 time=0, 10u (sfc 165)
    assert np.all(data[0, 1] == 1)   # date=20200101 time=0, 10v (sfc 166)
    assert np.all(data[1, 0] == 2)   # date=20200101 time=300, 10u (sfc 165)
    assert np.all(data[0, 2] == 31)  # date=20200101 time=0, u (pl 131) level=50
    assert np.all(data[0, 3] == 33)  # date=20200101 time=0, u (pl 131) level=100
    assert np.all(data[0, 4] == 32)  # date=20200101 time=0, v (pl 132) level=50
    assert np.all(data[0, 5] == 34)  # date=20200101 time=0, v (pl 132) level=100


def test_single_key_axes(read_only_fdb_setup_for_sfc_pl_example) -> None:
    """Separate date, time, param axes (no compound axes). Extension on axis 2."""
    builder = SimpleStoreBuilder(read_only_fdb_setup_for_sfc_pl_example)
    builder.add_part(
        SFC_REQUEST,
        [
            AxisDefinition(["date"], Chunking.SINGLE_VALUE),
            AxisDefinition(["time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    builder.add_part(
        PL_REQUEST,
        [
            AxisDefinition(["date"], Chunking.SINGLE_VALUE),
            AxisDefinition(["time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param", "levelist"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    builder.extend_on_axis(2)
    store = builder.build()

    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)

    assert data

    # data[date_idx, time_idx, param_idx]
    assert np.all(data[0, 0, 0] == 0)   # date=20200101, time=0, 10u (sfc 165)
    assert np.all(data[0, 0, 1] == 1)   # date=20200101, time=0, 10v (sfc 166)
    assert np.all(data[0, 0, 2] == 31)  # date=20200101, time=0, u (pl 131) level=50
    assert np.all(data[0, 0, 3] == 33)  # date=20200101, time=0, u (pl 131) level=100
    assert np.all(data[0, 0, 4] == 32)  # date=20200101, time=0, v (pl 132) level=50
    assert np.all(data[0, 0, 5] == 34)  # date=20200101, time=0, v (pl 132) level=100
    assert np.all(data[0, 1, 0] == 2)   # date=20200101, time=300, 10u (sfc 165)
    assert np.all(data[1, 0, 0] == 16)  # date=20200102, time=0, 10u (sfc 165)


def test_all_no_chunking(read_only_fdb_setup_for_sfc_pl_example) -> None:
    """Both axes NONE. Extension on axis 1. Single giant chunk."""
    builder = SimpleStoreBuilder(read_only_fdb_setup_for_sfc_pl_example)
    builder.add_part(
        SFC_REQUEST,
        [AxisDefinition(["date", "time"], Chunking.NONE), AxisDefinition(["param"], Chunking.NONE)],
        ExtractorType.GRIB,
    )
    builder.add_part(
        PL_REQUEST,
        [
            AxisDefinition(["date", "time"], Chunking.NONE),
            AxisDefinition(["param", "levelist"], Chunking.NONE),
        ],
        ExtractorType.GRIB,
    )
    builder.extend_on_axis(1)
    store = builder.build()

    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)

    assert data

    # data[date_time_idx, param_idx]
    assert np.all(data[0, 0] == 0)   # date=20200101 time=0, 10u (sfc 165)
    assert np.all(data[0, 1] == 1)   # date=20200101 time=0, 10v (sfc 166)
    assert np.all(data[0, 2] == 31)  # date=20200101 time=0, u (pl 131) level=50
    assert np.all(data[0, 3] == 33)  # date=20200101 time=0, u (pl 131) level=100
    assert np.all(data[0, 4] == 32)  # date=20200101 time=0, v (pl 132) level=50
    assert np.all(data[0, 5] == 34)  # date=20200101 time=0, v (pl 132) level=100



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
import yaml

from pyfdb._internal.pyfdb_internal import ConfigMapper


def test_config_mapper_wrong_type():
    with pytest.raises(ValueError):
        _ = ConfigMapper.to_json(0)


def test_config_mapper_none():
    json_config = ConfigMapper.to_json(None)
    assert json_config is None


def test_config_mapper_equality(read_only_fdb_setup):
    json_config_from_path = ConfigMapper.to_json(read_only_fdb_setup)
    json_config_from_str = ConfigMapper.to_json(read_only_fdb_setup.read_text())
    json_config_from_yaml = ConfigMapper.to_json(
        yaml.safe_load(read_only_fdb_setup.read_bytes())
    )

    assert json_config_from_path == json_config_from_str
    assert json_config_from_path == json_config_from_yaml



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from pathlib import Path

import pytest

from pyfdb import FDB


def assert_one_field(pyfdb: FDB, test_logger):
    selection = {
        "class": "rd",
        "expver": "xxxx",
        "stream": "oper",
        "date": "20191110",
        "time": "0000",
        "domain": "g",
        "type": "an",
        "levtype": "pl",
        "step": "0",
        "levelist": "300",
        "param": "138",
    }

    with pyfdb.retrieve(selection) as data_handle:
        assert data_handle

    list_iterator = pyfdb.list(selection)
    elements = list(list_iterator)

    for el in elements:
        test_logger.info(el)

    assert len(elements) == 1

    with pytest.raises(StopIteration) as _:
        next(list_iterator)


def test_archive_none(empty_fdb_setup, test_data_path, test_logger):
    fdb_config_path = empty_fdb_setup

    assert fdb_config_path

    fdb = FDB(fdb_config_path)
    filename: Path = test_data_path / "x138-300.grib"

    fdb.archive(filename.read_bytes())
    fdb.flush()

    assert_one_field(fdb, test_logger)


def test_archive_abitrary_data(empty_fdb_setup, test_logger):
    fdb_config_path = empty_fdb_setup

    assert fdb_config_path

    selection = {
        "class": "rd",
        "expver": "arb0",
        "stream": "oper",
        "date": "20191110",
        "time": "0000",
        "domain": "g",
        "type": "an",
        "levtype": "pl",
        "step": "0",
        "levelist": "300",
        "param": "138",
    }

    test_logger.debug("-----------ARCHIVE-------------------")

    with FDB(fdb_config_path) as fdb:
        fdb.archive(
            data=b"binary-data",
            identifier=selection,
        )

    test_logger.debug("-----------RETRIEVE-------------------")

    data_handle = fdb.retrieve(selection)
    file_content = data_handle.readall()
    assert file_content == b"binary-data"

    test_logger.debug("-----------CLOSE-------------------")


key_values = [
    ("stream", "oper/rd"),
    ("stream", ["oper", "rd"]),
]


@pytest.mark.parametrize("wrong_key_value", key_values)
def test_archive_abitrary_data_wrong_identifier(empty_fdb_setup, wrong_key_value):
    selection = {
        "class": "rd",
        "expver": "arb1",
        wrong_key_value[0]: wrong_key_value[1],
        "date": "20191110",
        "time": "0000",
        "domain": "g",
        "type": "an",
        "levtype": "pl",
        "step": "0",
        "levelist": "300",
        "param": "138",
    }

    with pytest.raises(ValueError), FDB(empty_fdb_setup) as fdb:
        fdb.archive(
            data=b"binary-data",
            identifier=selection,
        )


def test_archive_round_trip(empty_fdb_setup, test_data_path, test_logger):
    fdb = FDB(empty_fdb_setup)

    filename = test_data_path / "x138-300.grib"
    file_content = filename.read_bytes()

    test_logger.debug("-----------ARCHIVE-------------------")
    test_logger.debug(f"Read {len(file_content)} bytes")
    fdb.archive(file_content)
    fdb.flush()

    test_logger.debug("-----------RETRIEVE-------------------")

    # Those are the FDB keys of the GRIB file given above
    selection = {
        "class": "rd",
        "expver": "xxxx",
        "stream": "oper",
        "date": "20191110",
        "time": "0000",
        "domain": "g",
        "type": "an",
        "levtype": "pl",
        "step": "0",
        "levelist": "300",
        "param": "138",
    }

    with fdb.retrieve(selection) as data_handle:
        reread_file_content = data_handle.readall()
        test_logger.debug(f"Original size: {len(file_content)}")
        test_logger.debug(f"Reread size: {len(reread_file_content)}")
        assert file_content == reread_file_content

    list_iterator = fdb.list({})

    assert len(list(list_iterator)) == 1



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.

import argparse
import logging
import os
import sys

import findlibs

import pyfdb._internal as _internal

# INFO: This is in place because we currently can't (at runtime)
# tell which order the dependencies are in. This needs to be available in findlibs
DEPENDENCY_ORDER = ["eckit", "eccodes", "metkit", "fdb5"]
OPTIONAL_DEPENDENCIES = ["eccodes"]


def main():
    parser = argparse.ArgumentParser(description="pyfdb command line interface")
    parser.add_argument(
        "--print-home",
        action="store_true",
        help="Print the home directory of the fdb5 library",
    )
    parser.add_argument(
        "--print-home-deps",
        action="store_true",
        help="Print the home directories of all pyfdb dependencies",
    )
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        help="Enable DEBUG logging (default: INFO)",
    )
    args = parser.parse_args()

    logging.basicConfig(
        format="%(asctime)s | %(levelname)-6s | %(message)s",
        datefmt="%Y-%m-%d %H:%M:%S",
        level=logging.DEBUG if args.verbose else logging.INFO,
    )

    if not (args.print_home or args.print_home_deps):
        parser.print_help()
        sys.exit(2)

    def _lib_home(lib_path):
        lib_dir = os.path.dirname(os.path.realpath(lib_path))
        return (
            os.path.dirname(lib_dir)
            if os.path.basename(lib_dir) in ("lib", "lib64")
            else lib_dir
        )

    def _print_dep_path(lib, dependency_path, optional):
        missing = []
        if dependency_path is None:
            msg = f"\t{lib} [Optional]" if optional else f"\t{lib}"
            msg += ": not found by findlibs"
            missing.append(lib)
            if optional:
                logging.info(msg)
            else:
                logging.error(msg)
        else:
            msg = f"\t{lib} [Optional]" if optional else f"\t{lib}"
            msg += f": {_lib_home(dependency_path)}"
            logging.info(msg)

        return missing

    library_info_tuple = _internal.version_info()

    if args.print_home:
        dependency_path = findlibs.find("fdb5")
        if dependency_path is None:
            logging.error("fdb5 library not found by findlibs")
            sys.exit(1)
        for name, version, gitSha, path in library_info_tuple:
            if name == "fdb":
                logging.info(f"\t{name} {version} ({gitSha}) {path}")

    if args.print_home_deps:
        logging.info("Findlibs Environment:")
        for key, value in os.environ.items():
            if key.upper().startswith("FINDLIBS_DISABLE"):
                logging.info(f"\t{key: <15}: {value: <10}")

        logging.info("Findlibs Lookup")

        missing = []
        for lib in DEPENDENCY_ORDER:
            dependency_path = findlibs.find(lib)
            missing += _print_dep_path(
                lib, dependency_path, lib in OPTIONAL_DEPENDENCIES
            )

        logging.info("Dependency Versions:")

        for name, version, gitSha, path in library_info_tuple:
            logging.info(f"\t{name} {version} ({gitSha}) {path}")

        if missing:
            sys.exit(1)


if __name__ == "__main__":
    main()



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest

from pyfdb import FDB


def test_retrieve(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    data_handle = fdb.retrieve(selection)

    assert data_handle
    data_handle.open()
    assert data_handle.read(4) == b"GRIB"
    data_handle.close()


def test_retrieve_wildcard(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    with pytest.raises(TypeError):
        _ = fdb.retrieve({})


def test_retrieve_context_manager(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["167", "165", "166"],
        "time": "1800",
    }

    with fdb.retrieve(selection) as data_handle:
        assert data_handle
        assert data_handle.read(4) == b"GRIB"



import logging
from itertools import permutations, product

import pytest
import zarr

from z3fdb import (
    AxisDefinition,
    Chunking,
    SimpleStoreBuilder,
    ExtractorType,
)

logging.basicConfig(level=logging.DEBUG)


def test_access_pattern(read_only_fdb_pattern_setup) -> None:
    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/to/2020-01-03,"
        "levtype=sfc,"
        "step=0,"
        "param=165/166/167,"
        "time=0/to/21/by/6",
        [
            AxisDefinition(["date"], Chunking.SINGLE_VALUE),
            AxisDefinition(["time"], Chunking.SINGLE_VALUE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
            AxisDefinition(["step"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    store = builder.build()

    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)
    assert data

    # This is coming from the read_only_fdb_pattern_setup fixtures, the data is filled in there
    # in this pattern
    for date in range(0, 3):
        for time in range(0, 4):
            for param in range(0, 3):
                for step in range(0, 1):
                    index = step + param * 1 + time * 3 * 1 + date * 4 * 3 * 1
                    assert all(data[date, time, param, step] == index)


@pytest.mark.parametrize("index_permutation", permutations([0, 1, 2, 3]))
def test_access_pattern_shuffled_chunked(
    read_only_fdb_pattern_setup, index_permutation
) -> None:
    """Main idea is to test all permutations of the indices in a permuted way and check whether
    the received data aligns with the expected
    """
    axis = [
        AxisDefinition(["date"], Chunking.SINGLE_VALUE),
        AxisDefinition(["time"], Chunking.SINGLE_VALUE),
        AxisDefinition(["param"], Chunking.SINGLE_VALUE),
        AxisDefinition(["step"], Chunking.SINGLE_VALUE),
    ]
    axis_names = ["date", "time", "param", "step"]
    axis_permutation = [axis[i] for i in index_permutation]

    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/to/2020-01-03,"
        "levtype=sfc,"
        "step=0,"
        "param=165/166/167,"
        "time=0/to/21/by/6",
        axis_permutation,
        ExtractorType.GRIB,
    )
    store = builder.build()

    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)
    assert data

    logging.debug(f"Permutation: ({[axis_names[i] for i in index_permutation]})")

    def compute_value(date, time, param, step):
        return step + param * 1 + time * 3 * 1 + date * 4 * 3 * 1

    logging.debug(data.shape)
    logging.debug(data[:, 0, 0, 0])

    # This is coming from the read_only_fdb_pattern_setup fixtures, the data is filled in there
    # in this pattern
    for date in range(0, 3):
        for time in range(0, 4):
            for param in range(0, 3):
                for step in range(0, 1):
                    cur_index = [0, 0, 0, 0]

                    # Find the index the date gets mapped to, was index 0 in the non-permuted array
                    # Same for all other elements
                    cur_index[index_permutation.index(0)] = date
                    cur_index[index_permutation.index(1)] = time
                    cur_index[index_permutation.index(2)] = param
                    cur_index[index_permutation.index(3)] = step

                    value = compute_value(date=date, time=time, param=param, step=step)

                    logging.debug(f"({date}, {time}, {param}, {step})")
                    logging.debug(f"Permutation: {index_permutation}")
                    logging.debug(f"CurIndex: {cur_index}")
                    logging.debug(f"Value: {value}")
                    logging.debug(f"Data: {data[*cur_index]}")

                    assert all(data[*cur_index] == value)


test_data = product(
    permutations([0, 1, 2]),
    product([Chunking.SINGLE_VALUE, Chunking.NONE], repeat=3),
)


@pytest.mark.parametrize("index_permutation, chunked_permutations", test_data)
def test_access_pattern_shuffled_partially_chunked(
    read_only_fdb_pattern_setup, index_permutation, chunked_permutations
) -> None:
    """Main idea is to test all permutations of the indices in a permuted way and check whether
    the received data aligns with the expected. Same as test above but with additional merged axis and
    permuted chunked stati.
    """
    axis = [
        AxisDefinition(["date", "time"], chunked_permutations[0]),
        AxisDefinition(["param"], chunked_permutations[1]),
        AxisDefinition(["step"], chunked_permutations[2]),
    ]
    axis_names = ["datetime", "param", "step"]
    axis_permutation = [axis[i] for i in index_permutation]

    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/to/2020-01-03,"
        "levtype=sfc,"
        "step=0,"
        "param=165/166/167,"
        "time=0/to/21/by/6",
        axis_permutation,
        ExtractorType.GRIB,
    )
    store = builder.build()
    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)
    assert data

    logging.debug(f"Permutation: ({[axis_names[i] for i in index_permutation]})")
    logging.debug(f"Permutation chunked: ({[i for i in chunked_permutations]})")

    def compute_value(date, time, param, step):
        return step + param * 1 + time * 3 * 1 + date * 4 * 3 * 1

    logging.debug(f"Zarr array shape={data.shape}")
    logging.debug(f"Chunk shape={data.chunks}")
    logging.debug(data[:, 0, 0])

    # This is coming from the read_only_fdb_pattern_setup fixtures, the data is filled in there
    # in this pattern
    for date in range(0, 3):
        for time in range(0, 4):
            for param in range(0, 3):
                for step in range(0, 1):
                    datetime = date * 4 + time
                    # Find the index the date gets mapped to, was index 0 in the non-permuted array
                    # Same for all other elements
                    cur_index = [0, 0, 0]
                    cur_index[index_permutation.index(0)] = datetime
                    cur_index[index_permutation.index(1)] = param
                    cur_index[index_permutation.index(2)] = step

                    value = compute_value(date=date, time=time, param=param, step=step)

                    logging.debug(f"({date}, {time}, {param}, {step})")
                    logging.debug(f"Permutation: {index_permutation}")
                    logging.debug(f"CurIndex: {cur_index}")
                    logging.debug(f"Value: {value}")
                    logging.debug(f"Data: {data[*cur_index]}")

                    assert all(data[*cur_index] == value)


def test_access_pattern_non_chunked(read_only_fdb_pattern_setup) -> None:
    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/to/2020-01-03,"
        "levtype=sfc,"
        "step=0,"
        "param=165/166/167,"
        "time=0/to/21/by/6",
        [
            AxisDefinition(["date"], Chunking.NONE),
            AxisDefinition(["time"], Chunking.NONE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
            AxisDefinition(["step"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    store = builder.build()
    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)
    assert data

    for date in range(0, 3):
        for time in range(0, 4):
            for param in range(0, 3):
                for step in range(0, 1):
                    index = step + param * 1 + time * 3 * 1 + date * 4 * 3 * 1
                    assert all(data[date, time, param, step] == index)


def test_access_pattern_non_chunked_mixed(read_only_fdb_pattern_setup) -> None:
    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/to/2020-01-03,"
        "levtype=sfc,"
        "step=0,"
        "param=165/166/167,"
        "time=0/to/21/by/6",
        [
            AxisDefinition(["time"], Chunking.NONE),
            AxisDefinition(["date"], Chunking.NONE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
            AxisDefinition(["step"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    store = builder.build()
    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)
    assert data

    for date in range(0, 3):
        for time in range(0, 4):
            for param in range(0, 3):
                for step in range(0, 1):
                    index = step + param * 1 + time * 3 * 1 + date * 4 * 3 * 1
                    assert all(data[time, date, param, step] == index)


def test_access_pattern_merged_axis_non_chunked(read_only_fdb_pattern_setup) -> None:
    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/to/2020-01-03,"
        "levtype=sfc,"
        "step=0,"
        "param=165/166/167,"
        "time=0/to/21/by/6",
        [
            AxisDefinition(["date", "time"], Chunking.NONE),
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
            AxisDefinition(["step"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    store = builder.build()
    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)
    assert data

    for date in range(0, 3):
        for time in range(0, 4):
            for param in range(0, 3):
                for step in range(0, 1):
                    index = step + param * 1 + time * 3 * 1 + date * 4 * 3 * 1
                    datetime = time + date * 4
                    assert all(data[datetime, param, step] == index)


def test_access_pattern_merged_axis_non_chunked_switched_date_time(
    read_only_fdb_pattern_setup,
) -> None:
    builder = SimpleStoreBuilder(read_only_fdb_pattern_setup)
    builder.add_part(
        "type=an,"
        "class=ea,"
        "domain=g,"
        "expver=0001,"
        "stream=oper,"
        "date=2020-01-01/to/2020-01-03,"
        "levtype=sfc,"
        "step=0,"
        "param=165/166/167,"
        "time=0/to/21/by/6",
        [
            AxisDefinition(["param"], Chunking.SINGLE_VALUE),
            AxisDefinition(["time", "date"], Chunking.NONE),
            AxisDefinition(["step"], Chunking.SINGLE_VALUE),
        ],
        ExtractorType.GRIB,
    )
    store = builder.build()
    data = zarr.open_array(store, mode="r", zarr_format=3, use_consolidated=False)
    assert data

    for date in range(0, 3):
        for time in range(0, 4):
            for param in range(0, 3):
                for step in range(0, 1):
                    index = step + param * 1 + time * 3 * 1 + date * 4 * 3 * 1
                    datetime = date + time * 3
                    assert all(data[param, datetime, step] == index)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.

import logging
from collections.abc import Collection, Generator
from pathlib import Path
from typing import Any

from pyfdb._internal import (
    _FDB,
    Config,
    ConfigMapper,
    FDBToolRequest,
    init_bindings,
)
from pyfdb.pyfdb_iterator import (
    ControlElement,
    IndexAxis,
    ListElement,
    PurgeElement,
    StatsElement,
    StatusElement,
    WipeElement,
)
from pyfdb.pyfdb_type import (
    ControlAction,
    ControlIdentifier,
    DataHandle,
    MarsIdentifier,
    MarsSelection,
    UserInputMapper,
)


class FDB:
    def __init__(
        self,
        config: str | dict | Path | None = None,
        user_config: str | dict | Path | None = None,
    ) -> None:
        """
        Constructor for FDB object.

        Parameters
        ----------
        `config`: `str` | `dict` | `Path` | `None`, *optional*
            Config object for setting up the FDB. See Notes.
        `user_config`: `str` | `dict` | `Path` | `None`, *optional*
            Config object for setting up user specific options, e.g., enabling sub-TOCs. See Notes.

        Returns
        -------
        :returns: FDB object

        Note
        ----
        Every config parameter but is converted accordingly depending on its type:
            - `str` is used as a yaml representation to parse the config
            - `dict` is interpreted as hierarchical format to represent a config, see example
            - `Path` is interpreted as a location of the config and read as a YAML file
            - `None` is the fallback. The default config in `$FDB_HOME` is loaded

        Using a single PyFDB instance per individual threads is safe. Sharing the instances across threads isn't supported.
        However, the underlying FDB and its methods are thread-safe; the caller needs to be aware that flush acts on all archive calls,
        including archived messages from other threads. A call to flush will persist all archived messages regardless
        from which thread the message has been archived. In case the caller wants a finer control it is advised to
        instantiate one FDB object per thread to ensure only messages are flushed that have been archived on the same FDB
        object.

        Examples
        --------
        >>> fdb = pyfdb.FDB(fdb_config_path)
        >>> config = {
        ...     "type":"local",
        ...     "engine":"toc",
        ...     "schema":"<schema_path>",
        ...     "spaces":[
        ...         {
        ...             "handler":"Default",
        ...             "roots":[
        ...                 {"path": "<db_store_path>"},
        ...             ],
        ...         }
        ...     ],
        ... }
        >>> fdb = pyfdb.FDB(config)

        Or leveraging the context manager:

        >>> with pyfdb.FDB(fdb_config_path) as fdb:
        ...     # Call methods of fdb
        ...     pass
        """

        init_bindings()
        self.logger = logging.getLogger(__name__ + ".FDB")

        # Convert to JSON if set
        config = ConfigMapper.to_json(config)
        user_config = ConfigMapper.to_json(user_config)

        if config is not None and user_config is not None:
            internal_config = Config(config, user_config)
            self.FDB = _FDB(internal_config)
        elif config is not None:
            internal_config = Config(config, None)
            self.FDB = _FDB(internal_config)
        else:
            self.FDB = _FDB()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.flush()

    def archive(self, data: bytes, identifier: MarsIdentifier | None = None):
        """
        Archive binary data into the underlying FDB.
        In case an identifier is supplied, that identifier is used to archive the data.
        *No consistency checks are applied. The caller needs to ensure the provided identifier matches metadata present in data.*

        If no identifier is supplied, `data` is interpreted as GRIB data and the metadata is taken from
        the GRIB messages.

        In any case, the supplied or derived metadata needs to provide values for all required keys of the FDB schema.

        Parameters
        ----------
        `data`: `bytes`
            The binary data to be archived. If no key is provided this is interpreted by `eccodes` and may contain multiple GRIB messages.
        `identifier` : `Identifier` | None, optional
            A unique identifier for the archived data.
            - If provided, the data will be stored under this identifier.
            - If None, the data will be archived without an explicit identifier, metadata has to be derivable from the data, which is interpreted as GRIB data.

        Note
        ----
        Sometimes an identifier is also referred to as a Key.


        Returns
        -------
        None

        Examples
        --------
        >>> fdb = pyfdb.FDB()
        >>> filename = data_path / "x138-300.grib"
        >>> fdb.archive(data=filename.read_bytes()) # Archive
        >>> fdb.archive(identifier=Identifier([("key-1", "value-1")]), data=filename.read_bytes())
        >>> fdb.flush() # Sync the archive call
        """
        if identifier is None:
            self.FDB.archive(data, len(data))
        else:
            mapped_identifier = UserInputMapper.map_identifier_to_internal(identifier)
            self.FDB.archive(mapped_identifier, data, len(data))

    def flush(self):
        """
        Flush all buffers and close all data handles of the underlying FDB into a consistent DB state.
        *Always safe to call*

        Parameters
        ----------
        None

        Returns
        -------
        None

        Examples
        --------
        >>> fdb = pyfdb.FDB()
        >>> filename = data_path / "x138-300.grib"
        >>> fdb.archive(bytes=filename.read_bytes()) # Archive
        >>> fdb.flush() # Data is synced
        """
        self.FDB.flush()

    def retrieve(self, mars_selection: MarsSelection) -> DataHandle:
        """
        Retrieve data which is specified by a MARS selection.

        Parameters
        ----------
        `mars_selection`
            MARS selection which describes the data which should be retrieved

        Note
        ----
        The returned data handle doesn't guarantee the order of the GRIB messages.

        Returns
        -------
        DataHandle
            A data handle which contains unordered GRIB messages and can be read like a `BytesLike` object.

        Examples
        --------
        >>> mars_selection = {"key-1": "value-1", ...}
        >>> data_handle = pyfdb.retrieve(mars_selection)
        >>> data_handle.open()
        >>> data_handle.read(4)
        >>> data_handle.close()

        Or leveraging the context manager:

        >>> with pyfdb.retrieve(selection) as data_handle:
        >>>     assert data_handle
        >>>     assert data_handle.read(4) == b"GRIB"
        """
        if len(mars_selection) == 0:
            raise TypeError("FDB.retrieve: Wildcard selection aren't support for retrieving.")

        internal_mars_selection = UserInputMapper.map_selection_to_internal(mars_selection)
        return DataHandle(self.FDB.retrieve(internal_mars_selection), _internal=True)

    def list(
        self,
        mars_selection: MarsSelection,
        include_masked: bool = False,
        level: int = 3,
    ) -> Generator[ListElement, None, None]:
        """
        List data present at the underlying fdb archive and which can be retrieved.

        Parameters
        ----------
        `mars_selection` : `MarsSelection`
            A MARS selection which describes the data which can be listed. If `None` is given, all data will be listed.
        `include_masked` : bool, *optional*
            If True, the returned iterator lists masked data, if False the elements are unique.
        `level` : int [1-3], *optional*
            Specifies the FDB schema level of the elements which are matching the selection.
            A level of 1 means return a level 1 key (of the FDB schema) which is matching the MARS selection.

        Returns
        -------
        Generator[ListElement, None, None]
            A generator for `ListElement` describing FDB entries containing data of the MARS selection

        Note
        ----
        *this call lists masked elements if `include_masked` is `True`.*

        Examples
        --------
        >>> selection = {
        >>>     "type": "an",
        >>>     "class": "ea",
        >>>     "domain": "g",
        >>>     "expver": "0001",
        >>>     "stream": "oper",
        >>>     "date": "20200101",
        >>>     "levtype": "sfc",
        >>>     "step": "0",
        >>>     "time": "1800",
        >>> }
        >>> list_iterator = pyfdb.list(selection) # level == 3
        >>> elements = list(list_iterator)
        >>> print(elements[0])

        {class=ea,expver=0001,stream=oper,date=20200101,time=1800,domain=g}
        {type=an,levtype=sfc}
        {step=0,param=131},
        tocfieldlocation[uri=uri[scheme=file,name=<location>],offset=10732,length=10732,remapkey={}],
        length=10732,
        timestamp=176253515

        >>> list_iterator = pyfdb.list(selection, level=2)
        >>> elements = list(list_iterator)
        >>> print(elements[0])

        {class=ea,expver=0001,stream=oper,date=20200101,time=1800,domain=g}
        {type=an,levtype=sfc},
        length=0,
        timestamp=0

        >>> list_iterator = pyfdb.list(selection, level=1)
        >>> elements = list(list_iterator)
        >>> print(elements[0])

        {class=ea,expver=0001,stream=oper,date=20200101,time=1800,domain=g},
        length=0,
        timestamp=0
        """

        internal_mars_selection = UserInputMapper.map_selection_to_internal(mars_selection)
        fdb_tool_request = FDBToolRequest.from_internal_mars_selection(internal_mars_selection)
        iterator = self.FDB.list(fdb_tool_request.tool_request, not include_masked, level)
        for list_element in iterator:
            try:
                yield ListElement(list_element, _internal=True)
            except StopIteration:
                return

    def inspect(self, mars_selection: MarsSelection) -> Generator[ListElement, None, None]:
        """
        Inspects the content of the underlying FDB and returns a generator of list elements
        describing which field was part of the MARS selection.

        Parameters
        ----------
        `mars_selection` : `MarsSelection`
            An MARS selection for which the inspect should be executed

        Returns
        -------
        Generator[ListElement, None, None]
            A generator for `ListElement` describing FDB entries containing data of the MARS selection


        Examples
        --------
        >>> selection = {
        >>>         "type": "an",
        >>>         "class": "ea",
        >>>         "domain": "g",
        >>>         "expver": "0001",
        >>>         "stream": "oper",
        >>>         "date": "20200101",
        >>>         "levtype": "sfc",
        >>>         "step": "0",
        >>>         "param": "167",
        >>>         "time": "1800",
        >>>     }
        >>> list_iterator = pyfdb.inspect(selection)
        >>> elements = list(list_iterator) # single element in iterator
        >>> elements[0]
        {class=ea,expver=0001,stream=oper,date=20200101,time=1800,domain=g}
        {type=an,levtype=sfc}
        {param=167,step=0},
        TocFieldLocation[
            uri=URI[scheme=<location>],
            offset=0,
            length=10732,
            remapKey={}
        ],
        length=10732,
        timestamp=1762537447
        """
        internal_mars_selection = UserInputMapper.map_selection_to_internal(mars_selection)
        iterator = self.FDB.inspect(internal_mars_selection)
        for inspect_element in iterator:
            try:
                yield ListElement(inspect_element, _internal=True)
            except StopIteration:
                return

    def status(self, mars_selection: MarsSelection) -> Generator[StatusElement, None, None]:
        """
        List the status of all FDB entries with their control identifiers, e.g., whether a certain
        database was locked for retrieval.

        Parameters
        ----------
        `mars_selection` : `MarsSelection`
            An MARS selection which specifies the queried data

        Returns
        -------
        Generator[StatusElement, None, None]
            A generator for `StatusElement` describing FDB entries and their control identifier


        Examples
        --------
        >>> selection = {
        >>>         "type": "an",
        >>>         "class": "ea",
        >>>         "domain": "g",
        >>>     },
        >>> )
        >>> status_iterator = pyfdb.status(selection)
        >>> elements = list(status_iterator)
        >>> elements[0]
        StatusElement(
            control_identifiers=[],
            key={
                'class': ['ea'],
                'type': ['an'],
                'date': ['20200104'],
                'domain': ['g'],
                'expver': ['0001'],
                'stream': ['oper'],
                'time': ['2100']
                },
            location=/<path_to_root>/ea:0001:oper:20200104:2100:g
        )
        """
        internal_mars_selection = UserInputMapper.map_selection_to_internal(mars_selection)
        fdb_tool_request = FDBToolRequest.from_internal_mars_selection(internal_mars_selection)
        iterator = self.FDB.status(fdb_tool_request.tool_request)
        while True:
            try:
                yield StatusElement(next(iterator), _internal=True)
            except StopIteration:
                return

    def wipe(
        self,
        mars_selection: MarsSelection,
        doit: bool = False,
        porcelain: bool = False,
        unsafe_wipe_all: bool = False,
    ) -> Generator[WipeElement, None, None]:
        """
        Wipe data from the database.

        Delete FDB databases and the data therein contained. Use the passed
        selection to identify the database to delete. This is equivalent to a UNIX rm command.
        This function deletes either whole databases, or whole indexes within databases

        Parameters
        ----------
        `mars_selection` : `MarsSelection`
            An MARS selection which specifies the affected data
        `doit` : `bool`, *optional*
            If true the wipe command is executed, per default there are only dry-run
        `porcelain` : `bool`, *optional*
            Restricts the output to the wiped files
        `unsafe_wipe_all` : `bool`, *optional*
            Flag for disabling all security checks and force a wipe

        Returns
        -------
        Generator[WipeElement, None, None]
            A generator for `WipeElement`

        Note
        ----
        Wipe elements are not directly corresponding to the wiped files. This can be a cause for confusion.
        The individual wipe elements strings of the wipe output.

        Examples
        --------
        >>> fdb = pyfdb.FDB(fdb_config_path)
        >>> wipe_iterator = fdb.wipe({"class": "ea"})
        >>> wiped_elements = list(wipe_iterator)
        ...
        Toc files to delete:
        <path_to_database>/toc
        ...
        """
        internal_mars_selection = UserInputMapper.map_selection_to_internal(mars_selection)
        fdb_tool_request = FDBToolRequest.from_internal_mars_selection(internal_mars_selection)
        iterator = self.FDB.wipe(fdb_tool_request.tool_request, doit, porcelain, unsafe_wipe_all)
        for wipe_element in iterator:
            try:
                yield WipeElement(wipe_element, _internal=True)
            except StopIteration:
                return

    def purge(
        self,
        mars_selection: MarsSelection,
        doit: bool = False,
        porcelain: bool = False,
    ) -> Generator[PurgeElement, None, None]:
        """
        Remove duplicate data from the database.

        Purge duplicate entries from the database and remove the associated data if the data is owned and not adopted.
        Data in the FDB5 is immutable. It is masked, but not removed, when overwritten with new data using the same key.
        Masked data can no longer be accessed. Indexes and data files that only contains masked data may be removed.

        If an index refers to data that is not owned by the FDB (in particular data which has been adopted from an
        existing FDB5), this data will not be removed.

        Parameters
        ----------
        `mars_selection` : `MarsSelection`
            A MARS selection which describes the data which is purged.
        `doit` : `bool`, *optional*
            If true the wipe command is executed, per default there are only dry-run
        `porcelain` : `bool`, *optional*
            Restricts the output to the wiped files

        Returns
        -------
        Generator[PurgeElement, None, None]
            A generator for `PurgeElement`

        Examples
        --------
        >>> fdb = pyfdb.FDB(fdb_config_path)
        >>> purge_iterator = fdb.purge({"class": "ea"}), doit=True)
        >>> purged_elements = list(purge_iterator)
        >>> print(purged_elements[0])
        {class=ea,expver=0001,stream=oper,date=20200104,time=1800,domain=g}
        {type=an,levtype=sfc}
        {step=0,param=167},
        TocFieldLocation[
            uri=URI[
                scheme=file,
                name=<location>
            ],
            offset=32196,
            length=10732,
            remapKey={}
        ],
        length=10732,
        timestamp=176253976
        """
        internal_mars_selection = UserInputMapper.map_selection_to_internal(mars_selection)
        fdb_tool_request = FDBToolRequest.from_internal_mars_selection(internal_mars_selection)
        iterator = self.FDB.purge(fdb_tool_request.tool_request, doit, porcelain)
        for purge_element in iterator:
            try:
                yield PurgeElement(purge_element, _internal=True)
            except StopIteration:
                return

    def stats(self, mars_selection: MarsSelection) -> Generator[StatsElement, None, None]:
        """
        Print information about FDB databases, aggregating the
        information over all the databases visited into a final summary.

        Parameters
        ----------
        `mars_selection` : `MarsSelection`
            A MARS selection which specifies the affected data.

        Returns
        -------
        Generator[StatsElement, None, None]
            A generator for `StatsElement`

        Examples
        --------
        >>> fdb = pyfdb.FDB(fdb_config_path)
        >>> stats_iterator = fdb.stats(selection)
        >>> for el list(stats_iterator):
        >>>     print(el)
        Index Statistics:
        Fields                          : 3
        Size of fields                  : 32,196 (31.4414 Kbytes)
        Reacheable fields               : 3
        Reachable size                  : 32,196 (31.4414 Kbytes)

        DB Statistics:
        Databases                       : 1
        TOC records                     : 2
        Size of TOC files               : 2,048 (2 Kbytes)
        Size of schemas files           : 228 (228 bytes)
        TOC records                     : 2
        Owned data files                : 1
        Size of owned data files        : 32,196 (31.4414 Kbytes)
        Index files                     : 1
        Size of index files             : 131,072 (128 Kbytes)
        Size of TOC files               : 2,048 (2 Kbytes)
        Total owned size                : 165,544 (161.664 Kbytes)
        Total size                      : 165,544 (161.664 Kbytes)
        """
        internal_mars_selection = UserInputMapper.map_selection_to_internal(mars_selection)
        fdb_tool_request = FDBToolRequest.from_internal_mars_selection(internal_mars_selection)
        iterator = self.FDB.stats(fdb_tool_request.tool_request)
        for stats_element in iterator:
            try:
                yield StatsElement(stats_element, _internal=True)
            except StopIteration:
                return

    def control(
        self,
        mars_selection: MarsSelection,
        control_action: ControlAction,
        control_identifiers: Collection[ControlIdentifier],
    ) -> Generator[ControlElement, None, None]:
        """
        Enable certain features of FDB databases, e.g., disables or enables retrieving, list, etc.

        Parameters
        ----------
        `mars_selection` : `MarsSelection`
            A MARS selection which specifies the affected data.
        `control_action` : `ControlAction`
            Which action should be modified, e.g., ControlAction.RETRIEVE
        `control_identifiers` : `list[ControlIdentifier]`
            Should an action be enabled or disabled, e.g., ControlIdentifier.ENABLE or ControlIdentifier.DISABLE

        Returns
        -------
        Generator[ControlElement, None, None]
            A generator for `ControlElement`

        Note
        ----
        Disabling of an ControlAction, e.g., ControlAction.RETRIEVE leads to the creation
        of a `retrieve.lock` in the corresponding FDB database. This is true for all actions.
        The file is removed after the Action has been disabled.

        **It's important to consume the iterator, otherwise the lock file isn't deleted which
        can cause unexpected behavior. Also, due to internal reuse of databases, create a new FDB
        object before relying on the newly set control_identifier, to propagate the status.**

        Examples
        --------
        >>> fdb = pyfdb.FDB(fdb_config_path)
        >>> selection = {
        >>>         "class": "ea",
        >>>         "domain": "g",
        >>>         "expver": "0001",
        >>>         "stream": "oper",
        >>>         "date": "20200101",
        >>>         "time": "1800",
        >>> }
        >>> control_iterator = fdb.control(
        >>>     selection,
        >>>     ControlAction.DISABLE,
        >>>     [ControlIdentifier.RETRIEVE],
        >>> )
        >>> elements = list(control_iterator)
        >>> print(elements[0])
        ControlElement(
            control_identifiers=[RETRIEVE],
            key={
                'class': ['ea'],
                'date': ['20200104'],
                'domain': ['g'],
                'expver': ['0001'],
                'stream': ['oper'],
                'time': ['2100']
                },
            location=/<path_to_root>/ea:0001:oper:20200104:2100:g
        )
        """
        internal_mars_selection = UserInputMapper.map_selection_to_internal(mars_selection)
        fdb_tool_request = FDBToolRequest.from_internal_mars_selection(internal_mars_selection)
        raw_control_identifiers = [control_identifier._to_raw() for control_identifier in control_identifiers]
        iterator = self.FDB.control(
            fdb_tool_request.tool_request,
            control_action._to_raw(),
            raw_control_identifiers,
        )
        for control_element in iterator:
            try:
                yield ControlElement(control_element, _internal=True)
            except StopIteration:
                return

    def axes(self, mars_selection: MarsSelection, level: int = 3) -> IndexAxis:
        """
        Return the 'axes' and their extent of a MARS selection for a given level of the schema in
        an IndexAxis object.

        If a key isn't specified the entire extent (all values) are returned.

        Parameters
        ----------
        `mars_selection` : `MarsSelection`
            A MARS selection which specifies the affected data.
        `level` : int [1-3], *optional*
            Level of the FDB Schema. Only keys of the given level are returned.

        Returns
        -------
        IndexAxis
            A map containing Key-Value pairs of the axes and their extent

        Examples
        --------
        >>> fdb = pyfdb.FDB(fdb_config_path)
        >>> selection = {
        ...         "type": "an",
        ...         "class": "ea",
        ...         "domain": "g",
        ...         "expver": "0001",
        ...         "stream": "oper",
        ...         "levtype": "sfc",
        ...         "step": "0",
        ...         "time": "1800",
        ... }
        >>> index_axis: IndexAxis = fdb.axes(selection) # level == 3
        >>> for k, v in index_axis.items():
        ...     print(f"k={k} \t| v={v}")
        k=class    | v=['ea']
        k=date     | v=['20200101', '20200102', '20200103', '20200104']
        k=domain   | v=['g']
        k=expver   | v=['0001']
        k=levelist | v=['']
        k=levtype  | v=['sfc']
        k=param    | v=['131', '132', '167']
        k=step     | v=['0']
        k=stream   | v=['oper']
        k=time     | v=['1800']
        k=type     | v=['an']
        """
        internal_mars_selection = UserInputMapper.map_selection_to_internal(mars_selection)
        fdb_tool_request = FDBToolRequest.from_internal_mars_selection(internal_mars_selection)
        return IndexAxis(self.FDB.axes(fdb_tool_request.tool_request, level))

    def enabled(self, control_identifier: ControlIdentifier) -> bool:
        """
        Check whether a specific control identifier is enabled

        Parameters
        ----------
        `control_identifier` : `ControlIdentifier`
            A given control identifier

        Returns
        -------
        `bool`
            `True` if the given control identifier is set, `False` otherwise.

        Examples
        --------
        >>> fdb_config = yaml.safe_load(fdb_config_path)
        >>> fdb_config["writable"] = False
        >>> fdb = pyfdb.FDB(fdb_config)
        >>> fdb.enabled(ControlIdentifier.NONE) # == True
        >>> fdb.enabled(ControlIdentifier.LIST) # == True
        >>> fdb.enabled(ControlIdentifier.RETRIEVE) # == True
        >>> fdb.enabled(ControlIdentifier.ARCHIVE) # == False, default True
        >>> fdb.enabled(ControlIdentifier.WIPE) # == False, default True
        >>> fdb.enabled(ControlIdentifier.UNIQUEROOT) # == True

        """
        return self.FDB.enabled(control_identifier._to_raw())

    def dirty(self):
        """
        Return whether a flush of the FDB is needed, for example if data was archived since the last flush.

        Parameters
        ----------
        None

        Returns
        -------
        `bool`
            `True` if an archive happened and a flush is needed, `False` otherwise.


        Examples
        --------
        >>> fdb = FDB(fdb_config_file)
        >>> filename = <data_path>
        >>> fdb.archive(open(filename, "rb").read())
        >>> fdb.dirty()                         # == True
        >>> fdb.flush()
        >>> fdb.dirty()                         # == False

        """
        return self.FDB.dirty()

    def config(self) -> tuple[dict[str, Any], dict[str, Any]]:
        """
        Return the system and user configuration of the underlying FDB.

        Parameters
        ----------
        None

        Returns
        -------
        `tuple[dict[str, Any], dict[str, Any]]`
            Python dictionaries describing the system and user configuration


        Examples
        --------
        >>> fdb = FDB(config_file)
        >>> system_config, user_config = fdb.config()
        >>> print(system_config)
        >>> print(user_config)

        """
        system_config = ConfigMapper.from_json(self.FDB.config().json())
        user_config = ConfigMapper.from_json(self.FDB.config().userConfig().json())

        return system_config, user_config

    def __repr__(self) -> str:
        return repr(self.FDB)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest
import yaml

from pyfdb import FDB


def test_initialization():
    fdb = FDB()
    assert fdb


def test_fdb_config_default():
    assert FDB()


def test_fdb_config_wrong_type():
    with pytest.raises(
        ValueError, match="Config: Unknown config type, must be str, dict or Path"
    ):
        fdb = FDB(0)

        assert fdb


def test_fdb_config_fixture(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)
    assert fdb


def test_fdb_config_equality(read_only_fdb_setup):
    pyfdb_config_str = FDB(read_only_fdb_setup)
    assert pyfdb_config_str

    pyfdb_config_path = FDB(read_only_fdb_setup)
    assert pyfdb_config_path

    config_dict = yaml.safe_load(read_only_fdb_setup.read_bytes())
    pyfdb_config_dict = FDB(config_dict)
    assert pyfdb_config_dict

    print(pyfdb_config_str)
    print(pyfdb_config_path)
    print(pyfdb_config_dict)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
    }

    elements_str = list(pyfdb_config_str.status(selection))
    elements_path = list(pyfdb_config_path.status(selection))
    elements_dict = list(pyfdb_config_dict.status(selection))

    print(elements_str)
    print(elements_dict)
    print(elements_str)

    assert all(
        x == y == z for (x, y, z) in zip(elements_str, elements_path, elements_dict)
    )


def test_fdb_user_config(read_only_fdb_setup):
    user_config_str = r"""---
        type: local
        engine: toc
        useSubToc: true
        spaces:
        - roots:
          - path: "/a/path/is/something"
    """

    fdb = FDB(read_only_fdb_setup, user_config_str)
    assert fdb

    print("Check for user config propagation:")
    print(fdb.config())

    system_config, user_config = fdb.config()
    assert "useSubToc" in user_config
    assert user_config["useSubToc"] is True

    fdb_no_user_config = FDB(read_only_fdb_setup)
    print(fdb_no_user_config.config())
    print("Check for empty user config:")
    system_config_no_user_config, user_config_no_user_config = (
        fdb_no_user_config.config()
    )

    assert system_config == system_config_no_user_config
    assert user_config
    assert len(user_config_no_user_config) == 0


def test_fdb_print():
    fdb = FDB()

    assert fdb
    print(fdb)


def test_fdb_context():
    with FDB() as fdb:
        assert fdb
        print(fdb)



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

from pathlib import Path

import pytest
import yaml

from pyfdb import FDB, ControlAction, ControlIdentifier
from pyfdb._internal import _ControlAction, _ControlIdentifier
from pyfdb.pyfdb_type import UserInputMapper


def test_control_action_values():
    assert ControlAction.NONE.value == _ControlAction.NONE.value
    assert ControlAction.DISABLE.value == _ControlAction.DISABLE.value
    assert ControlAction.ENABLE.value == _ControlAction.ENABLE.value


def test_control_identifier_values():
    assert _ControlIdentifier.NONE.value == ControlIdentifier.NONE.value
    assert _ControlIdentifier.LIST.value == ControlIdentifier.LIST
    assert _ControlIdentifier.RETRIEVE.value == ControlIdentifier.RETRIEVE
    assert _ControlIdentifier.ARCHIVE.value == ControlIdentifier.ARCHIVE
    assert _ControlIdentifier.WIPE.value == ControlIdentifier.WIPE
    assert _ControlIdentifier.UNIQUEROOT.value == ControlIdentifier.UNIQUEROOT


def test_control_identifier_string():
    assert str(ControlIdentifier.NONE) == "NONE"
    assert str(ControlIdentifier.LIST) == "LIST"
    assert str(ControlIdentifier.RETRIEVE) == "RETRIEVE"
    assert str(ControlIdentifier.ARCHIVE) == "ARCHIVE"
    assert str(ControlIdentifier.WIPE) == "WIPE"
    assert str(ControlIdentifier.UNIQUEROOT) == "UNIQUEROOT"

    assert str(ControlIdentifier.NONE | ControlIdentifier.LIST) == "LIST"
    assert str(ControlIdentifier.LIST | ControlIdentifier.RETRIEVE) == "LIST|RETRIEVE"
    assert str(ControlIdentifier.RETRIEVE | ControlIdentifier.NONE) == "RETRIEVE"
    assert (
        str(
            ControlIdentifier.RETRIEVE | ControlIdentifier.NONE | ControlIdentifier.LIST
        )
        == "LIST|RETRIEVE"
    )


def test_control_action_string():
    assert str(ControlAction.NONE) == "NONE"
    assert str(ControlAction.ENABLE) == "ENABLE"
    assert str(ControlAction.DISABLE) == "DISABLE"


def test_control_lock_retrieve(read_only_fdb_setup, function_tmp):
    fdb = FDB(read_only_fdb_setup)

    print("Retrieve without lock")
    data_handle = fdb.retrieve(
        {
            "type": "an",
            "class": "ea",
            "domain": "g",
            "expver": "0001",
            "stream": "oper",
            "date": "20200101",
            "levtype": "sfc",
            "step": "0",
            "param": "167/165/166",
            "time": "1800",
        }
    )
    assert data_handle
    data_handle.open()
    assert data_handle.read(4) == b"GRIB"
    data_handle.close()

    assert not (
        function_tmp / "db_store" / "ea:0001:oper:20200101:1800:g" / "retrieve.lock"
    ).exists()

    print("Locking database for retrieve")

    selection = {
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "time": "1800",
    }

    control_iterator = fdb.control(
        selection,
        ControlAction.DISABLE,
        [ControlIdentifier.RETRIEVE],
    )
    assert control_iterator

    elements = []

    for el in control_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 1

    assert (
        function_tmp / "db_store" / "ea:0001:oper:20200101:1800:g" / "retrieve.lock"
    ).exists()

    print("Retrieve with lock")
    # Needed because reuse of databases in the retrieve path is using a cached db
    # object which contains the first (non-locked) status of the db.
    # This can be removed as soon as https://jira.ecmwf.int/browse/FDB-613 is fixed.
    fdb = FDB(read_only_fdb_setup)

    data_handle = fdb.retrieve(
        {
            "type": "an",
            "class": "ea",
            "domain": "g",
            "expver": "0001",
            "stream": "oper",
            "date": "20200101",
            "levtype": "sfc",
            "step": "0",
            "param": "167/165/166",
            "time": "1800",
        },
    )
    assert data_handle
    data_handle.open()
    assert data_handle.read(4) == b""
    data_handle.close()


def test_control_lock_list(read_only_fdb_setup, function_tmp):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "time": "1800",
    }

    print("List without lock")
    list_iterator = fdb.list(selection)
    elements = list(list_iterator)
    assert len(elements) == 3

    print("Locking database for listing")
    control_iterator = fdb.control(
        selection, ControlAction.DISABLE, [ControlIdentifier.LIST]
    )
    assert control_iterator

    elements = []

    for el in control_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 1

    assert (
        function_tmp / "db_store" / "ea:0001:oper:20200101:1800:g" / "list.lock"
    ).exists()

    print("List with lock. Expecting 0 elements.")
    list_iterator = fdb.list(selection)
    elements = list(list_iterator)
    assert len(elements) == 0

    print("Unlocking database for listing")
    control_iterator = fdb.control(
        selection, ControlAction.ENABLE, [ControlIdentifier.LIST]
    )
    assert control_iterator

    elements = list(control_iterator)
    assert len(elements) > 0

    # Correct behaviour is to "not see the data after locking"
    assert not (
        function_tmp / "db_store" / "ea:0001:oper:20200101:1800:g" / "list.lock"
    ).exists()

    list_iterator = fdb.list(selection)
    elements = list(list_iterator)
    assert len(elements) == 3


def test_control_lock_archive(read_only_fdb_setup, build_grib_messages, function_tmp):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "time": "1800",
    }

    print("Lock the database for archiving")
    control_iterator = fdb.control(
        selection, ControlAction.DISABLE, [ControlIdentifier.ARCHIVE]
    )

    elements = []

    for el in control_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 1

    assert (
        function_tmp / "db_store" / "ea:0001:oper:20200101:1800:g" / "archive.lock"
    ).exists()

    print("Try archiving")
    with pytest.raises(
        Exception, match=" matched for archived is LOCKED against archiving"
    ):
        fdb.archive(build_grib_messages.read_bytes())
        fdb.flush()

    print("Unlock the database for archiving")
    control_iterator = fdb.control(
        selection, ControlAction.ENABLE, [ControlIdentifier.ARCHIVE]
    )
    assert control_iterator

    elements = []

    for el in control_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 1

    assert not (
        function_tmp / "db_store" / "ea:0001:oper:20200101:1800:g" / "archive.lock"
    ).exists()

    print("Try archiving")
    fdb.archive(build_grib_messages.read_bytes())
    fdb.flush()
    print("Success")


def test_control_lock_archive_status(
    read_only_fdb_setup, build_grib_messages, function_tmp, test_logger
):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "time": "1800",
    }

    test_logger.info("Lock the database for archiving")
    control_iterator = fdb.control(
        selection, ControlAction.DISABLE, [ControlIdentifier.ARCHIVE]
    )
    assert control_iterator

    elements = []

    for el in control_iterator:
        test_logger.debug(el)
        elements.append(el)

    assert len(elements) == 1

    test_logger.debug("--------- OUTPUT STATUS -----------")

    status_iterator = fdb.status(selection)

    elements = list(status_iterator)

    for el in elements:
        test_logger.debug(el)

    assert (
        function_tmp / "db_store" / "ea:0001:oper:20200101:1800:g" / "archive.lock"
    ).exists()

    test_logger.info("Try archiving")
    with pytest.raises(
        Exception, match=" matched for archived is LOCKED against archiving"
    ):
        fdb.archive(build_grib_messages.read_bytes())
        fdb.flush()

    test_logger.info("Unlock the database for archiving")
    control_iterator = fdb.control(
        selection, ControlAction.ENABLE, [ControlIdentifier.ARCHIVE]
    )
    assert control_iterator

    elements = []

    for el in control_iterator:
        test_logger.debug(el)
        elements.append(el)

    assert len(elements) == 1

    assert not (
        function_tmp / "db_store" / "ea:0001:oper:20200101:1800:g" / "archive.lock"
    ).exists()

    test_logger.info("Try archiving")
    fdb.archive(build_grib_messages.read_bytes())
    fdb.flush()
    test_logger.info("Success")


def test_control_lock_wipe(read_only_fdb_setup, function_tmp, build_grib_messages):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "time": "1800",
    }

    print("Archive same data as the FDB was setup with in the Fixture")
    fdb.archive(build_grib_messages.read_bytes())
    fdb.flush()

    print("Lock the database for wiping")
    control_iterator = fdb.control(
        selection, ControlAction.DISABLE, [ControlIdentifier.WIPE]
    )
    assert control_iterator

    elements = []

    for el in control_iterator:
        print(el)
        elements.append(el)

    assert len(elements) == 1
    assert (
        function_tmp / "db_store" / "ea:0001:oper:20200101:1800:g" / "wipe.lock"
    ).exists()

    print("Try Wipe")
    wipe_iterator = fdb.wipe(selection, doit=True)

    elements = []

    with pytest.raises(RuntimeError):
        for el in wipe_iterator:
            elements.append(el)

    assert len(elements) == 0

    print("Unlock the database for wiping")
    control_iterator = fdb.control(
        selection, ControlAction.ENABLE, [ControlIdentifier.WIPE]
    )
    assert control_iterator

    elements = []

    for el in control_iterator:
        print(el)
        elements.append(el)

    assert len(elements) > 0

    assert not (
        function_tmp / "db_store" / "ea:0001:oper:20200101:1800:g" / "wipe.lock"
    ).exists()

    print("Try Wipe")
    fdb.wipe(selection, doit=True)
    fdb.flush()

    print("Success")


def test_enabled_per_default(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    assert fdb.enabled(ControlIdentifier.NONE) is True
    assert fdb.enabled(ControlIdentifier.LIST) is True
    assert fdb.enabled(ControlIdentifier.RETRIEVE) is True
    assert fdb.enabled(ControlIdentifier.ARCHIVE) is True
    assert fdb.enabled(ControlIdentifier.WIPE) is True
    assert fdb.enabled(ControlIdentifier.UNIQUEROOT) is True


def test_disabled_writable(read_only_fdb_setup):
    fdb_config_path = read_only_fdb_setup

    assert fdb_config_path

    fdb_config = yaml.safe_load(fdb_config_path.read_text())
    fdb_config["writable"] = False

    fdb = FDB(yaml.dump(fdb_config))

    assert fdb.enabled(ControlIdentifier.NONE) is True
    assert fdb.enabled(ControlIdentifier.LIST) is True
    assert fdb.enabled(ControlIdentifier.RETRIEVE) is True
    assert fdb.enabled(ControlIdentifier.ARCHIVE) is False
    assert fdb.enabled(ControlIdentifier.WIPE) is False
    assert fdb.enabled(ControlIdentifier.UNIQUEROOT) is True


def test_disabled_visitable(read_only_fdb_setup):
    fdb_config = yaml.safe_load(read_only_fdb_setup.read_text())
    fdb_config["visitable"] = False

    fdb = FDB(yaml.dump(fdb_config))

    assert fdb.enabled(ControlIdentifier.NONE) is True
    assert fdb.enabled(ControlIdentifier.LIST) is False
    assert fdb.enabled(ControlIdentifier.RETRIEVE) is False
    assert fdb.enabled(ControlIdentifier.ARCHIVE) is True
    assert fdb.enabled(ControlIdentifier.WIPE) is True
    assert fdb.enabled(ControlIdentifier.UNIQUEROOT) is True


def test_disabled_visitable_writeable(read_only_fdb_setup):
    fdb_config = yaml.safe_load(read_only_fdb_setup.read_text())
    fdb_config["visitable"] = False
    fdb_config["writable"] = False

    fdb = FDB(yaml.dump(fdb_config))

    assert fdb.enabled(ControlIdentifier.NONE) is True
    assert fdb.enabled(ControlIdentifier.LIST) is False
    assert fdb.enabled(ControlIdentifier.RETRIEVE) is False
    assert fdb.enabled(ControlIdentifier.ARCHIVE) is False
    assert fdb.enabled(ControlIdentifier.WIPE) is False
    assert fdb.enabled(ControlIdentifier.UNIQUEROOT) is True


def test_needs_flush(empty_fdb_setup, test_data_path):
    fdb = FDB(empty_fdb_setup)

    filename: Path = test_data_path / "x138-300.grib"

    fdb.archive(filename.read_bytes())
    assert fdb.dirty() is True
    fdb.flush()
    assert fdb.dirty() is False


def test_control_element_key(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "time": "1800",
    }

    print("List without lock")
    list_iterator = fdb.list(selection)
    elements = list(list_iterator)
    assert len(elements) == 3

    print("Locking database for listing")
    control_iterator = fdb.control(
        selection, ControlAction.DISABLE, [ControlIdentifier.LIST]
    )
    assert control_iterator

    elements = []

    for el in control_iterator:
        print(el)
        elements.append(el.key())

    assert len(elements) == 1
    print(f"Element key: {elements[0]}")
    print(f"Element key type: {type(elements[0])}")
    assert elements[0] == UserInputMapper.map_selection_to_internal(selection)


def test_control_element_control_identifiers(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "time": "1800",
    }

    print("List without lock")
    list_iterator = fdb.list(selection)
    elements = list(list_iterator)
    assert len(elements) == 3

    control_identifier_list = [
        ControlIdentifier.LIST,
        ControlIdentifier.RETRIEVE,
        ControlIdentifier.ARCHIVE,
        ControlIdentifier.WIPE,
        ControlIdentifier.UNIQUEROOT,
    ]

    print("Locking database for all control actions")
    control_iterator = fdb.control(
        selection, ControlAction.DISABLE, control_identifiers=control_identifier_list
    )
    assert control_iterator

    elements = []

    for el in control_iterator:
        print(el)
        elements.append(el.controlIdentifiers())

    assert len(elements) == 1
    assert (
        len(elements[0]) == 5
    )  # There should be 5 control identifiers for the first database
    assert elements[0] == control_identifier_list
    print(f"Element control identifiers: {elements[0]}")
    print(f"Element control identifiers type: {type(elements[0])}")

    print("Unlocking database for all control actions")
    control_iterator = fdb.control(
        selection, ControlAction.ENABLE, control_identifiers=control_identifier_list
    )
    assert control_iterator

    elements = [el.controlIdentifiers() for el in control_iterator]

    assert len(elements) == 1
    assert len(elements[0]) == 0
    print(f"Element control identifiers: {elements[0]}")
    print(f"Element control identifiers type: {type(elements[0])}")


def test_status_element_eq_control_element(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "time": "1800",
    }

    print("List without lock")
    list_iterator = fdb.list(selection)
    elements = list(list_iterator)
    assert len(elements) == 3

    control_identifier_list = [
        ControlIdentifier.LIST,
        ControlIdentifier.RETRIEVE,
        ControlIdentifier.ARCHIVE,
        ControlIdentifier.WIPE,
        ControlIdentifier.UNIQUEROOT,
    ]

    print("Locking all databases for all control actions")

    # Control results for all databases
    control_iterator = fdb.control(
        {}, ControlAction.DISABLE, control_identifiers=control_identifier_list
    )
    assert control_iterator

    control_elements = list(control_iterator)

    # Status for all databases
    status_iterator = fdb.status({})
    assert status_iterator

    status_elements = list(status_iterator)
    assert control_elements == status_elements

    for i in range(len(control_elements)):
        print(f"Control Element: \n {control_elements[i]}")
        print(f"Status Element: \n {status_elements[i]}")

    # Check (cross-verify) that consecutive elements are different
    for i in range(len(control_elements) - 1):
        assert control_elements[i + 1] != status_elements[i]



# (C) Copyright 2025- ECMWF.
# .
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation nor
# does it submit to any jurisdiction.

# libfdb5.so and dependencies have to be loaded prior to importing
# pyfdb
import findlibs

findlibs.load("fdb5")

from pyfdb._internal.pyfdb_internal import (
    ConfigMapper,
    FDBToolRequest,
)
from pyfdb_bindings.pyfdb_bindings import (
    FDB as _FDB,
)
from pyfdb_bindings.pyfdb_bindings import (
    URI as _URI,
)
from pyfdb_bindings.pyfdb_bindings import (
    Config,
    ControlElement,
    IndexAxis,
    ListElement,
    StatsElement,
    WipeElement,
    WipeElementType,
    PurgeElement,
    init_bindings,
    version_info,
)
from pyfdb_bindings.pyfdb_bindings import (
    ControlAction as _ControlAction,
)
from pyfdb_bindings.pyfdb_bindings import (
    ControlIdentifier as _ControlIdentifier,
)
from pyfdb_bindings.pyfdb_bindings import (
    DataHandle as _DataHandle,
)

__all__ = [
    "init_bindings",
    "version_info",
    "_DataHandle",
    "_URI",
    "_FDB",
    "Config",
    "ConfigMapper",
    "FDBToolRequest",
    "InternalMarsSelection",
    "_ControlAction",
    "_ControlIdentifier",
    "ListElement",
    "StatsElement",
    "ControlElement",
    "WipeElement",
    "WipeElementType",
    "PurgeElement",
    "IndexAxis",
    "InternalMarsSelection",
    "MarsSelection",
    "UserInputMapper",
]



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import pytest

from pyfdb.pyfdb import FDB


def test_list_backward_compat_list(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["131", "132", "167"],
        "time": "1800",
    }

    elements = list(fdb.list(selection))

    for el in elements:
        print(el)

    assert len(elements) == 3


def test_list_backward_compat_list_of_list_str(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)
    # Next case should err, as this asks for a parameter "131/132/167" and
    # is wrapped in a list.
    selection_inconsistent = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": ["131/132/167"],
        "time": "1800",
    }

    with pytest.raises(RuntimeError, match="UserError"):
        _ = list(fdb.list(selection_inconsistent))


def test_list_backward_compat_str(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)
    selection_backward_comp = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101",
        "levtype": "sfc",
        "step": "0",
        "param": "131/132/167",
        "time": "1800",
    }

    elements = list(fdb.list(selection_backward_comp))

    for el in elements:
        print(el)

    assert len(elements) == 3


def test_list_backward_compat_to_list(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": ["20200101", "to", "20200103"],
        "levtype": "sfc",
        "step": "0",
        "param": ["131", "132", "167"],
        "time": "1800",
    }

    elements = list(fdb.list(selection))

    for el in elements:
        print(el)

    assert len(elements) == 9


def test_list_backward_compat_to_str(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101/to/20200103",
        "levtype": "sfc",
        "step": "0",
        "param": ["131", "132", "167"],
        "time": "1800",
    }

    elements = list(fdb.list(selection))

    for el in elements:
        print(el)

    assert len(elements) == 9


def test_list_backward_compat_list_of_str_list(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    # Next case should err, as this asks for a date "20200101/to/20200102" and
    # is wrapped in a list.
    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": ["20200101/to/20200103"],
        "levtype": "sfc",
        "step": "0",
        "param": ["131", "132", "167"],
        "time": "1800",
    }

    with pytest.raises(RuntimeError, match="UserError: Bad value"):
        _ = list(fdb.list(selection))


def test_list_backward_compat_to_by_list(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": ["20200101", "to", "20200104", "by", "1"],
        "levtype": "sfc",
        "step": "0",
        "param": ["131", "132", "167"],
        "time": "1800",
    }

    elements = list(fdb.list(selection))

    for el in elements:
        print(el)

    assert len(elements) == 12


def test_list_backward_compat_to_by_str(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "20200101/to/20200104/by/1",
        "levtype": "sfc",
        "step": "0",
        "param": ["131", "132", "167"],
        "time": "1800",
    }

    elements = list(fdb.list(selection))

    for el in elements:
        print(el)

    assert len(elements) == 12


def test_list_backward_compat_to_by_list_of_str(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)

    # Next case should err, as this asks for a date "20200101/to/20200102" and
    # is wrapped in a list.
    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": ["20200101/to/20200104/by/1"],
        "levtype": "sfc",
        "step": "0",
        "param": ["131", "132", "167"],
        "time": "1800",
    }

    with pytest.raises(RuntimeError, match="UserError: Bad value"):
        _ = list(fdb.list(selection))


def test_list_backward_compat_list_relative_dates_list(
    empty_fdb_setup, build_grib_messages_relative_dates
):
    fdb = FDB(empty_fdb_setup)

    with fdb:
        fdb.archive(build_grib_messages_relative_dates.read_bytes())

    elements = list(fdb.list({}))

    for el in elements:
        print(el)

    selection = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": ["-2", "-1", "0"],
        "levtype": "sfc",
        "step": "0",
        "param": ["131", "132", "167"],
        "time": "1800",
    }

    elements = list(fdb.list(selection))

    for el in elements:
        print(el)

    assert len(elements) == 9


def test_list_backward_compat_list_relative_dates_list_of_str(
    empty_fdb_setup, build_grib_messages_relative_dates
):
    fdb = FDB(empty_fdb_setup)

    with fdb:
        fdb.archive(build_grib_messages_relative_dates.read_bytes())

    # Next case should err, as this asks for a parameter "131/132/167" and
    # is wrapped in a list.
    selection_inconsistent = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": ["-2/-1/0"],
        "levtype": "sfc",
        "step": "0",
        "param": ["131/132/167"],
        "time": "1800",
    }

    with pytest.raises(RuntimeError, match="UserError"):
        _ = list(fdb.list(selection_inconsistent))


def test_list_backward_compat_list_relative_dates_str(
    empty_fdb_setup, build_grib_messages_relative_dates
):
    fdb = FDB(empty_fdb_setup)

    with fdb:
        fdb.archive(build_grib_messages_relative_dates.read_bytes())

    selection_backward_comp = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "-2/-1/0",
        "levtype": "sfc",
        "step": "0",
        "param": "131/132/167",
        "time": "1800",
    }

    elements = list(fdb.list(selection_backward_comp))

    for el in elements:
        print(el)

    assert len(elements) == 9


def test_list_backward_compat_list_relative_dates_to_str(
    empty_fdb_setup, build_grib_messages_relative_dates
):
    fdb = FDB(empty_fdb_setup)

    with fdb:
        fdb.archive(build_grib_messages_relative_dates.read_bytes())
    selection_backward_comp = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "-2/to/0",
        "levtype": "sfc",
        "step": "0",
        "param": "131/132/167",
        "time": "1800",
    }

    elements = list(fdb.list(selection_backward_comp))

    for el in elements:
        print(el)

    assert len(elements) == 9


def test_list_backward_compat_list_relative_dates_to_by_str(
    empty_fdb_setup, build_grib_messages_relative_dates
):
    fdb = FDB(empty_fdb_setup)

    with fdb:
        fdb.archive(build_grib_messages_relative_dates.read_bytes())
    selection_backward_comp = {
        "type": "an",
        "class": "ea",
        "domain": "g",
        "expver": "0001",
        "stream": "oper",
        "date": "-2/to/0/by/1",
        "levtype": "sfc",
        "step": "0",
        "param": "131/132/167",
        "time": "1800",
    }

    elements = list(fdb.list(selection_backward_comp))

    for el in elements:
        print(el)

    assert len(elements) == 9



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import yaml

from pyfdb import FDB


def test_initialization_config_fixture(read_only_fdb_setup):
    fdb = FDB(read_only_fdb_setup)
    assert fdb


def test_initialization_config_equality(read_only_fdb_setup):
    pyfdb_config_str = FDB(read_only_fdb_setup.read_text())
    assert pyfdb_config_str

    pyfdb_config_path = FDB(read_only_fdb_setup)
    assert pyfdb_config_path

    config_dict = yaml.safe_load(read_only_fdb_setup.read_bytes())
    pyfdb_config_dict = FDB(config_dict)
    assert pyfdb_config_dict

    assert pyfdb_config_str.config() == pyfdb_config_path.config()
    assert pyfdb_config_path.config() == pyfdb_config_dict.config()


def test_initialization_config_different(read_only_fdb_setup):
    fdb_user_config_str_a = """
    type: local
    engine: toc
    useSubToc: true
    spaces:
    - roots:
      - path: "/a/path/is/something"
    """

    fdb_a = FDB(read_only_fdb_setup, fdb_user_config_str_a)

    fdb_user_config_str_b = """
    type: local
    engine: toc
    useSubToc: true
    spaces:
    - roots:
      - path: "/b/path/is/something"
    """

    fdb_b = FDB(read_only_fdb_setup, fdb_user_config_str_b)

    assert fdb_a.config() != fdb_b.config()


def test_initialization_config_system_and_user(read_only_fdb_setup):
    fdb_user_config_str = """
    type: local
    engine: toc
    useSubToc: true
    spaces:
    - roots:
      - path: "/a/path/is/something"
    """

    fdb = FDB(read_only_fdb_setup, fdb_user_config_str)
    assert fdb

    system_config, user_config = fdb.config()

    assert system_config
    assert user_config["useSubToc"] is True


def test_initialization_config_system_and_user_printing(read_only_fdb_setup):
    fdb_user_config_str = """
    type: local
    engine: toc
    useSubToc: true
    spaces:
    - roots:
      - path: "/a/path/is/something"
    """

    fdb = FDB(read_only_fdb_setup, fdb_user_config_str)
    assert fdb

    system_config, user_config = fdb.config()

    assert "db_store" in repr(system_config)
    assert "'useSubToc': True" in repr(user_config)


def test_initialization_config_system_and_user_round_trip(read_only_fdb_setup):
    fdb_user_config_str = """
    type: local
    engine: toc
    useSubToc: true
    spaces:
    - roots:
      - path: "/a/path/is/something"
    """

    fdb = FDB(read_only_fdb_setup, fdb_user_config_str)
    assert fdb

    system_config, user_config = fdb.config()

    fdb2 = FDB(system_config, user_config)
    assert fdb2

    system_config_2, user_config_2 = fdb2.config()
    assert system_config == system_config_2
    assert user_config == user_config_2



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.


import datetime

import pytest

from pyfdb._internal.pyfdb_internal import FDBToolRequest


def test_from_internal_selection():
    today_date = datetime.date.today()
    yesterday = int((today_date - datetime.timedelta(days=1)).strftime("%Y%m%d"))
    fdb_tool_request = FDBToolRequest.from_internal_mars_selection({"date": ["-1"]})

    print(fdb_tool_request)
    str_repr = str(fdb_tool_request)

    assert "retrieve" in str_repr
    assert f"date={yesterday}" in str_repr


def test_from_selection():
    # Internal representation is not respected
    with pytest.raises(ValueError, match="Expecting collection of values"):
        _ = FDBToolRequest.from_internal_mars_selection({"date": "-1"})


def test_from_selection_all():
    fdb_tool_request = FDBToolRequest.from_internal_mars_selection({})

    print(fdb_tool_request)
    str_repr = str(fdb_tool_request)

    assert "ALL" in str_repr



# (C) Copyright 2025- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import logging
import pathlib
from typing import Generator

import datetime

import pytest
import itertools
import shutil
import yaml
import git

import eccodes as ec

import pyfdb

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


@pytest.fixture
def test_logger():
    return logger


@pytest.fixture(scope="function")
def data_path() -> pathlib.Path:
    """
    Provides path to test data
    """
    path = pathlib.Path(__file__).resolve().parent / "data"
    assert path.exists()
    return path


@pytest.fixture(scope="function")
def test_data_path() -> pathlib.Path:
    """
    Provides path to test data
    """

    git_repo = git.Repo(__file__, search_parent_directories=True)
    git_root = pathlib.Path(git_repo.git.rev_parse("--show-toplevel"))
    path = git_root / "tests" / "pyfdb" / "data"
    assert path.exists()
    return path


@pytest.fixture(scope="session")
def session_tmp(tmp_path_factory) -> Generator[pathlib.Path, None, None]:
    """Temporary folder for use in a single session"""
    tmp_dir = tmp_path_factory.mktemp("session_data")
    yield tmp_dir
    shutil.rmtree(str(tmp_dir))


@pytest.fixture(scope="function")
def function_tmp(tmp_path_factory) -> Generator[pathlib.Path, None, None]:
    """Temporary folder for use in a single function"""
    tmp_function_dir = tmp_path_factory.mktemp("test_data", numbered=True)
    yield tmp_function_dir
    shutil.rmtree(str(tmp_function_dir))


@pytest.fixture(scope="function")
def build_grib_messages(data_path, session_tmp) -> pathlib.Path:
    """This fixture is setting up a certain set of GRIB files by using a template GRIB file.
    For every triple of dates, times and parameters as stated below a new field is created
    and written to a file. This file path is returned at the end of the method to be usable
    in other fixtures.
    """
    template_grib = data_path / "template.grib"
    assert template_grib.is_file()
    template_grib_fd = open(template_grib, "rb")
    gid = ec.codes_grib_new_from_file(template_grib_fd)
    template_grib_fd.close()
    count_data_points = int(ec.codes_get(gid, "numberOfDataPoints"))
    count_values = int(ec.codes_get(gid, "numberOfValues"))
    count_missing = int(ec.codes_get(gid, "numberOfMissing"))

    # This only supports messages without missing datapoints
    assert count_data_points == count_values
    assert count_missing == 0

    # Set common keys / data "pattern"
    ec.codes_set_string(gid, "type", "an")
    ec.codes_set_string(gid, "class", "ea")
    ec.codes_set_string(gid, "expver", "0001")
    ec.codes_set_string(gid, "stream", "oper")
    ec.codes_set_string(gid, "levtype", "sfc")
    ec.codes_set_values(gid, list(range(0, count_values)))

    dates = [20200101, 20200102, 20200103, 20200104]
    times = [0, 300, 600, 900, 1200, 1500, 1800, 2100]
    parameters = [167, 131, 132]

    messages = session_tmp / "test_data.grib"
    with open(messages, "wb") as out:
        for date, time, parameter in itertools.product(dates, times, parameters):
            ec.codes_set(gid, "date", date)
            ec.codes_set(gid, "time", time)
            ec.codes_set(gid, "paramId", parameter)
            ec.codes_write(gid, out)

    ec.codes_release(gid)
    return messages


@pytest.fixture(scope="function", autouse=False)
def empty_fdb_setup(data_path, session_tmp, function_tmp) -> pathlib.Path:
    """
    Creates a FDB setup in this tests temp directory.
    This setup can be shared between tests
    """
    schema_path_src = data_path / "schema"
    assert schema_path_src.is_file()
    schema_path = session_tmp / "schema"
    shutil.copy(schema_path_src, schema_path)

    db_store_path = session_tmp / function_tmp / "db_store"
    db_store_path.mkdir()
    fdb_config = dict(
        type="local",
        engine="toc",
        schema=str(schema_path),
        spaces=[
            dict(
                handler="Default",
                roots=[
                    {"path": str(db_store_path)},
                ],
            )
        ],
    )
    fdb_config_str = yaml.dump(fdb_config)
    fdb_config_path = session_tmp / "fdb_config.yaml"
    fdb_config_path.write_text(fdb_config_str)
    return fdb_config_path


@pytest.fixture(scope="function", autouse=False)
def read_only_fdb_setup(empty_fdb_setup, build_grib_messages) -> pathlib.Path:
    """
    Creates a FDB setup in this tests temp directory.
    Test FDB currently reads all grib files in `tests/data`
    This setup can be shared between tests as we will only read
    data from this FDB
    """

    logger.debug("***********READ ONLY FDB SETUP*******************")
    with pyfdb.FDB(empty_fdb_setup) as fdb:
        fdb.archive(build_grib_messages.read_bytes())
        fdb.flush()

    return empty_fdb_setup


@pytest.fixture(scope="function", autouse=False)
def read_write_fdb_setup(empty_fdb_setup, build_grib_messages) -> pathlib.Path:
    """
    Creates a FDB setup in this tests temp directory.
    This setup can be shared between tests as we will only read
    data from this FDB
    """
    with pyfdb.FDB(empty_fdb_setup) as fdb:
        fdb.archive(build_grib_messages.read_bytes())
        fdb.flush()
    return empty_fdb_setup


@pytest.fixture(scope="function")
def build_grib_messages_relative_dates(data_path, session_tmp) -> pathlib.Path:
    """This fixture is setting up a certain set of GRIB files by using a template GRIB file.
    For every triple of dates, times and parameters as stated below a new field is created
    and written to a file. This file path is returned at the end of the method to be usable
    in other fixtures.

    Note:
    This is the same as the build_grib_messages fixture but sets the dates to be relative
    to today.
    """
    template_grib = data_path / "template.grib"
    assert template_grib.is_file()
    template_grib_fd = open(template_grib, "rb")
    gid = ec.codes_grib_new_from_file(template_grib_fd)
    template_grib_fd.close()
    count_data_points = int(ec.codes_get(gid, "numberOfDataPoints"))
    count_values = int(ec.codes_get(gid, "numberOfValues"))
    count_missing = int(ec.codes_get(gid, "numberOfMissing"))

    # This only supports messages without missing datapoints
    assert count_data_points == count_values
    assert count_missing == 0

    # Set common keys / data "pattern"
    ec.codes_set_string(gid, "type", "an")
    ec.codes_set_string(gid, "class", "ea")
    ec.codes_set_string(gid, "expver", "0001")
    ec.codes_set_string(gid, "stream", "oper")
    ec.codes_set_string(gid, "levtype", "sfc")
    ec.codes_set_values(gid, list(range(0, count_values)))

    today_date = datetime.date.today()

    today = int(today_date.strftime("%Y%m%d"))
    yesterday = int((today_date - datetime.timedelta(days=1)).strftime("%Y%m%d"))
    before_yesterday = int((today_date - datetime.timedelta(days=2)).strftime("%Y%m%d"))
    before_before_yesterday = int(
        (today_date - datetime.timedelta(days=3)).strftime("%Y%m%d")
    )

    dates = [before_before_yesterday, before_yesterday, yesterday, today]
    times = [0, 300, 600, 900, 1200, 1500, 1800, 2100]
    parameters = [167, 131, 132]

    messages = session_tmp / "test_data.grib"
    with open(messages, "wb") as out:
        for date, time, parameter in itertools.product(dates, times, parameters):
            ec.codes_set(gid, "date", date)
            ec.codes_set(gid, "time", time)
            ec.codes_set(gid, "paramId", parameter)
            ec.codes_write(gid, out)

    ec.codes_release(gid)
    return messages
